local Cartographer = {}
_G.Cartographer3 = Cartographer

Cartographer.version = GetAddOnMetadata("Cartographer3", "Version")
Cartographer.date = nil

local db

local L = {}
setmetatable(L, {__index=function(self, key)
	self[key] = key
	return key
end})

local BLIZZARD_MAP_WIDTH = 1002
local BLIZZARD_MAP_HEIGHT_TO_WIDTH_RATIO = 2/3
local BLIZZARD_MINIMAP_TILE_YARD_SIZE = 533 + 1/3

local CITIES
local INDOOR_CITIES
local CONTINENT_DATA
local ZONE_DATA = {}
local LOCALIZED_ZONE_TO_TEXTURE = {}
local MINIMUM_ZOOM, MAXIMUM_ZOOM, ZOOM_STEP
local DEFAULT_ZOOM_TO_MINIMAP_TEXTURE
local DEFAULT_MAPFRAME_WIDTH, DEFAULT_MAPFRAME_HEIGHT
local MAPFRAME_MINRESIZE_WIDTH, MAPFRAME_MINRESIZE_HEIGHT
local DEFAULT_MAPVIEW_BACKGROUND_COLOR
local YARDS_PER_PIXEL
local ITERATIVE_PROCESS_STEP_SECONDS
local ZOOM_TIME, MANUAL_ZOOM_TIME
local FADE_TIME
local COORDINATE_SEPARATOR

local NORMAL_STATUS_COLOR, COMBAT_COLOR, DEAD_COLOR, INACTIVE_COLOR

local OVERLAY_DATA

local TEXTURE_DATA, INSTANCE_TEXTURE_DATA

local INSTANCE_ROTATIONS, INSTANCE_FLOOR_SPLITS

function Cartographer:SetConstants(data)
	self.SetConstants = nil
	CITIES = data.CITIES
	INDOOR_CITIES = data.INDOOR_CITIES
	CONTINENT_DATA = data.CONTINENT_DATA
	MINIMUM_ZOOM = data.MINIMUM_ZOOM
	ZOOM_STEP = data.ZOOM_STEP
	DEFAULT_ZOOM_TO_MINIMAP_TEXTURE = data.DEFAULT_ZOOM_TO_MINIMAP_TEXTURE
	DEFAULT_MAPFRAME_WIDTH = data.DEFAULT_MAPFRAME_WIDTH
	DEFAULT_MAPFRAME_HEIGHT = data.DEFAULT_MAPFRAME_HEIGHT
	MAPFRAME_MINRESIZE_WIDTH = data.MAPFRAME_MINRESIZE_WIDTH
	MAPFRAME_MINRESIZE_HEIGHT = data.MAPFRAME_MINRESIZE_HEIGHT
	DEFAULT_MAPVIEW_BACKGROUND_COLOR = data.DEFAULT_MAPVIEW_BACKGROUND_COLOR
	YARDS_PER_PIXEL = data.YARDS_PER_PIXEL
	ITERATIVE_PROCESS_STEP_SECONDS = data.ITERATIVE_PROCESS_STEP_SECONDS
	ZOOM_TIME = data.ZOOM_TIME
	MANUAL_ZOOM_TIME = data.MANUAL_ZOOM_TIME
	FADE_TIME = data.FADE_TIME
	COORDINATE_SEPARATOR = data.COORDINATE_SEPARATOR
	INSTANCE_ROTATIONS = data.INSTANCE_ROTATIONS
	INSTANCE_FLOOR_SPLITS = data.INSTANCE_FLOOR_SPLITS
	
	-- minimap width in pixels / yards at closest radius * yards per pixel
	MAXIMUM_ZOOM = 140 / (133 + 1/3) * YARDS_PER_PIXEL
	-- able to see twice what the minimap can show
	if not DEFAULT_ZOOM_TO_MINIMAP_TEXTURE then
		DEFAULT_ZOOM_TO_MINIMAP_TEXTURE = MAXIMUM_ZOOM / 7
	end
	
	NORMAL_STATUS_COLOR = data.NORMAL_STATUS_COLOR
	COMBAT_COLOR = data.COMBAT_COLOR
	DEAD_COLOR = data.DEAD_COLOR
	INACTIVE_COLOR = data.INACTIVE_COLOR
end

function Cartographer:SetOverlayData(data)
	self.SetOverlayData = nil
	OVERLAY_DATA = data
end

function Cartographer:SetTextureData(data)
	self.SetTextureData = nil
	TEXTURE_DATA = data
end

function Cartographer:SetInstanceTextureData(data)
	self.SetInstanceTextureData = nil
	INSTANCE_TEXTURE_DATA = data
end

local CONTINENT_DATA__TEXTURE, CONTINENT_DATA__INDEX, CONTINENT_DATA__LOCALIZED_NAME, CONTINENT_DATA__MAP_CENTER, CONTINENT_DATA__MAP_MIDDLE, CONTINENT_DATA__WIDTH, CONTINENT_DATA__HEIGHT, CONTINENT_DATA__SHOWN_LEFT, CONTINENT_DATA__SHOWN_RIGHT, CONTINENT_DATA__SHOWN_BOTTOM, CONTINENT_DATA__SHOWN_TOP, CONTINENT_DATA__ZONES, CONTINENT_DATA__MINIMAP_TEXTURE_OFFSET_X, CONTINENT_DATA__MINIMAP_TEXTURE_OFFSET_Y, CONTINENT_DATA__MINIMAP_TEXTURE_BACKGROUND, CONTINENT_DATA__MINIMAP_TEXTURE_CENTER_X, CONTINENT_DATA__MINIMAP_TEXTURE_CENTER_Y, CONTINENT_DATA__EXTRA_MINIMAP_TEXTURE_AREAS, CONTINENT_DATA__RECT_LEFT, CONTINENT_DATA__RECT_RIGHT, CONTINENT_DATA__RECT_TOP, CONTINENT_DATA__RECT_BOTTOM = 1, 2, 3, 4, 5, 6, 7, 8, 9, 10, 11, 12, 13, 14, 15, 16, 17, 18, 19, 20, 21, 22

local ZONE_DATA__CONTINENT_NUM, ZONE_DATA__INDEX, ZONE_DATA__LOCALIZED_NAME, ZONE_DATA__MAP_WIDTH, ZONE_DATA__MAP_HEIGHT, ZONE_DATA__MAP_LEFT, ZONE_DATA__MAP_BOTTOM, ZONE_DATA__SHOWN_MAP_LEFT, ZONE_DATA__SHOWN_MAP_BOTTOM, ZONE_DATA__HIGHLIGHT_LEFT, ZONE_DATA__HIGHLIGHT_RIGHT, ZONE_DATA__HIGHLIGHT_BOTTOM, ZONE_DATA__HIGHLIGHT_TOP = 1, 2, 3, 4, 5, 6, 7, 8, 9, 10, 11, 12, 13

local GetPlayerRotation
do
	local rotateMinimap = GetCVar("rotateMinimap") == "1"
	hooksecurefunc("SetCVar", function(key, value)
		if key == "rotateMinimap" then
			rotateMinimap = value == "1"
		end
	end)
	
	local MINIMAP_ARROW
	do
		local function f(...)
			local best
			for i = 1, select('#', ...) do
				local v = select(i, ...)
				if not v:GetName() and v:GetObjectType() == "Model" then
					best = v
				end
			end
			return best
		end
		MINIMAP_ARROW = f(Minimap:GetChildren())
	end

	function GetPlayerRotation()
		if rotateMinimap then
			return MiniMapCompassRing:GetFacing()
		else
			return -MINIMAP_ARROW:GetFacing()
		end
	end
end

local mapHolder, mapFrame, scrollframe, mapView

local WorldMapFrame = _G.WorldMapFrame
WorldMapTooltip:SetParent(UIParent)
WorldMapTooltip:SetFrameStrata("TOOLTIP")
local currentOwner, currentPoint
local old_WorldMapTooltip_SetOwner = WorldMapTooltip.SetOwner
function WorldMapTooltip.SetOwner(this, owner, point)
	-- do a SetOwner hack since if you set it to the owner inside the mapView, then there are scaling issues and it would be stuck inside the scroll child. This way, it gives the same effect as the author intended.
	currentOwner = owner
	currentPoint = point
	old_WorldMapTooltip_SetOwner(this, mapHolder, "ANCHOR_NONE")
	this:ClearAllPoints()
	if point == "ANCHOR_TOPRIGHT" then
		this:SetPoint("BOTTOMRIGHT", currentOwner, "TOPRIGHT")
	elseif point == "ANCHOR_RIGHT" then
		this:SetPoint("BOTTOMLEFT", currentOwner, "TOPRIGHT")
	elseif point == "ANCHOR_BOTTOMRIGHT" then
		this:SetPoint("TOPLEFT", currentOwner, "BOTTOMRIGHT")
	elseif point == "ANCHOR_TOPLEFT" then
		this:SetPoint("BOTTOMLEFT", currentOwner, "TOPLEFT")
	elseif point == "ANCHOR_LEFT" then
		this:SetPoint("BOTTOMRIGHT", currentOwner, "TOPLEFT")
	elseif point == "ANCHOR_BOTTOMLEFT" then
		this:SetPoint("TOPRIGHT", currentOwner, "BOTTOMLEFT")
	elseif point == "ANCHOR_CURSOR" then
		old_WorldMapTooltip_SetOwner(this, mapHolder, "ANCHOR_CURSOR")
	elseif point == "ANCHOR_PRESERVE" then
		this:SetPoint("BOTTOM", UIParent, "BOTTOM", 0, 80)
	end
end
function WorldMapTooltip.IsOwned(this, owner)
	return currentOwner == owner
end
local function HideWorldMapFrames(...)
	for i = 1, select('#', ...) do
		local v = select(i, ...)
		local name = v:GetName()
		if name and name:match("^WorldMap") then
			v:Hide()
			v:SetParent(nil)
		end
	end
end	
UIPanelWindows["WorldMapFrame"] = nil
WorldMapFrame:SetAttribute("UIPanelLayout-enabled", false)
WorldMapFrame:SetFrameStrata("MEDIUM")
WorldMapFrame:EnableMouse(nil)
WorldMapFrame:EnableMouseWheel(nil)
WorldMapFrame:ClearAllPoints()
WorldMapFrame:SetWidth(1024)
WorldMapFrame:SetHeight(768)
HideWorldMapFrames(_G.WorldMapDetailFrame:GetRegions())
HideWorldMapFrames(_G.WorldMapButton:GetChildren())
_G.WorldMapButton:EnableMouse(false)
_G.WorldMapButton:SetScript("OnUpdate", nil)
function _G.WorldMapButton_OnClick()
	-- hopefully I don't destroy another addon's hooks here, but *shrug*
end
_G.WorldMapButton:SetScript("OnEvent", nil)
_G.WorldMapButton:UnregisterAllEvents()
WorldMapFrame:SetScript("OnShow", nil)
WorldMapFrame:SetScript("OnHide", nil)
WorldMapFrame:SetScript("OnEvent", nil)
WorldMapFrame:UnregisterAllEvents()
WorldMapFrame:SetScript("OnUpdate", nil)
WorldMapFrame:SetScript("OnKeyDown", nil)
_G.WorldMapDetailFrame:ClearAllPoints()
_G.WorldMapButton:SetParent(_G.WorldMapDetailFrame)
WorldMapFrame:SetPoint("CENTER", _G.WorldMapDetailFrame, "CENTER", 2, 17) -- get relatively in the same position
function WorldMapFrame:GetScale()
	return self:GetEffectiveScale()
end
_G.WorldMapPositioningGuide:SetPoint("CENTER", _G.WorldMapDetailFrame, "CENTER", 2, 17)

local function IsMouseHovering(frame)
	if not frame:IsVisible() then
		return false
	end
	local x, y = GetCursorPosition()
	local scale = frame:GetEffectiveScale()
	x, y = x/scale, y/scale
	
	local left, bottom, width, height = frame:GetRect()
	
	return x >= left and y >= bottom and x <= left+width and y <= bottom+height
end

local discoveredOverlays = {}

local CURRENT_MAP_TEXTURE = GetMapInfo()
local CURRENT_CONTINENT_NUM = GetCurrentMapContinent()
local CURRENT_ZONE_DATA, CURRENT_CONTINENT_DATA
local overlaysChecked = {}
function Cartographer:WORLD_MAP_UPDATE()
	CURRENT_MAP_TEXTURE = GetMapInfo()
	CURRENT_CONTINENT_NUM = GetCurrentMapContinent()
	if not CURRENT_MAP_TEXTURE then
		if CURRENT_CONTINENT_NUM == -1 then
			CURRENT_MAP_TEXTURE = "Cosmic"
		else
			CURRENT_MAP_TEXTURE = "World"
		end
	end
	CURRENT_ZONE_DATA = ZONE_DATA[CURRENT_MAP_TEXTURE]
	CURRENT_CONTINENT_DATA = CONTINENT_DATA[CURRENT_CONTINENT_NUM]
	
	if CURRENT_ZONE_DATA then
		local scale = CURRENT_ZONE_DATA[ZONE_DATA__MAP_WIDTH] / BLIZZARD_MAP_WIDTH
		_G.WorldMapDetailFrame:SetScale(scale)
		_G.WorldMapButton:SetScale(scale)
		WorldMapFrame:SetScale(scale)
		_G.WorldMapPositioningGuide:SetScale(scale)
		_G.WorldMapDetailFrame:SetPoint("BOTTOMLEFT", mapView, "CENTER", CURRENT_ZONE_DATA[ZONE_DATA__MAP_LEFT] / scale, CURRENT_ZONE_DATA[ZONE_DATA__MAP_BOTTOM] / scale)
	elseif CURRENT_CONTINENT_DATA and CURRENT_CONTINENT_DATA[1] then
		local scale = CURRENT_CONTINENT_DATA[CONTINENT_DATA__WIDTH] / BLIZZARD_MAP_WIDTH
		_G.WorldMapDetailFrame:SetScale(scale)
		_G.WorldMapButton:SetScale(scale)
		WorldMapFrame:SetScale(scale)
		_G.WorldMapPositioningGuide:SetScale(scale)
		_G.WorldMapDetailFrame:SetPoint("CENTER", mapView, "CENTER", CURRENT_CONTINENT_DATA[CONTINENT_DATA__MAP_CENTER] / scale, CURRENT_CONTINENT_DATA[CONTINENT_DATA__MAP_MIDDLE] / scale)
	end
	
	if mapView then
		Cartographer:PLAYER_DEAD()
	end
	
	if overlaysChecked[CURRENT_MAP_TEXTURE] then
		return
	end
	overlaysChecked[CURRENT_MAP_TEXTURE] = true
	
	local numOverlays = GetNumMapOverlays()
	local len = #CURRENT_MAP_TEXTURE + 21 -- #([=[Interface\WorldMap\]=] .. CURRENT_MAP_TEXTURE .. [=[\]=])
	for i = 1, numOverlays do
		local tname, tw, th, ofx, ofy = GetMapOverlayInfo(i)
		tname = tname:sub(len)
		if not discoveredOverlays[tname] then
			local num = tw + th * 1024 + ofx * 1048576 + ofy * 1073741824
			if num ~= 0 and num ~= 131200 and tname ~= "" and tname:lower() ~= "pixelfix" then
				discoveredOverlays[tname] = num
				local OVERLAY_DATA__CURRENT_MAP_TEXTURE = OVERLAY_DATA[CURRENT_MAP_TEXTURE]
				if not OVERLAY_DATA__CURRENT_MAP_TEXTURE then
					OVERLAY_DATA__CURRENT_MAP_TEXTURE = {}
					OVERLAY_DATA[CURRENT_MAP_TEXTURE] = OVERLAY_DATA__CURRENT_MAP_TEXTURE
				end
				if OVERLAY_DATA__CURRENT_MAP_TEXTURE[tname] ~= num then
					OVERLAY_DATA__CURRENT_MAP_TEXTURE[tname] = num
					local Cartographer3_NewOverlaysDB = _G.Cartographer3_NewOverlaysDB
					if type(Cartographer3_NewOverlaysDB) ~= "table" then
						Cartographer3_NewOverlaysDB = {}
						_G.Cartographer3_NewOverlaysDB = Cartographer3_NewOverlaysDB
					end
					if not Cartographer3_NewOverlaysDB[CURRENT_MAP_TEXTURE] then
						Cartographer3_NewOverlaysDB[CURRENT_MAP_TEXTURE] = {}
					end
					Cartographer3_NewOverlaysDB[CURRENT_MAP_TEXTURE][tname] = num
				end
			end
		end
	end
end

local MATCH_ERR_ZONE_EXPLORED = "^" .. ERR_ZONE_EXPLORED:gsub("%%s", ".+") .. "$"
function Cartographer:UI_INFO_MESSAGE(event, message)
	if message:match(MATCH_ERR_ZONE_EXPLORED) then
		local fileName = LOCALIZED_ZONE_TO_TEXTURE[GetRealZoneText()]
		if fileName then
			overlaysChecked[fileName] = nil
		end
	end
end

local timers = {}
local timers_whenNotShownAlso = {}
local function addTimer(func, whenNotShownAlso)
	timers[func] = true
	if whenNotShownAlso then
		timers_whenNotShownAlso[func] = true
	end
end
local function removeTimer(func)
	timers[func] = nil
	timers_whenNotShownAlso[func] = nil
end

local function getScaledCursorPosition()
	local x, y = GetCursorPosition()
	local scale = UIParent:GetEffectiveScale()
	return x/scale, y/scale
end
Cartographer.getScaledCursorPosition = getScaledCursorPosition

local zoneSearchData = {} -- data in here is split up by continent and then into a 10x10 grid of possible zones, in the order they should be searched (by city and then by area)
local function convertUniverseCoordinateToZonePosition(x, y)
	for continentNum, currentContinentData in ipairs(CONTINENT_DATA) do
		local continentLeft = currentContinentData[CONTINENT_DATA__SHOWN_LEFT]
		local continentRight = currentContinentData[CONTINENT_DATA__SHOWN_RIGHT]
		local continentBottom = currentContinentData[CONTINENT_DATA__SHOWN_BOTTOM]
		local continentTop = currentContinentData[CONTINENT_DATA__SHOWN_TOP]
		if x >= continentLeft and x <= continentRight and y >= continentBottom and y <= continentTop then
			local x_grid = math.floor((x - continentLeft) / (continentRight - continentLeft) * 10)
			local y_grid = math.floor((y - continentBottom) / (continentTop - continentBottom) * 10)
			local num = x_grid*10 + y_grid
			local searchData = zoneSearchData[continentNum][num]
			local bestZone
			if searchData then
				for _, zoneTexture in ipairs(searchData) do
					local zoneData = ZONE_DATA[zoneTexture]
				
					if x >= zoneData[ZONE_DATA__HIGHLIGHT_LEFT] and x <= zoneData[ZONE_DATA__HIGHLIGHT_RIGHT] and y >= zoneData[ZONE_DATA__HIGHLIGHT_BOTTOM] and y <= zoneData[ZONE_DATA__HIGHLIGHT_TOP] then
						bestZone = zoneTexture
						break
					end
				end
			end
			
			if bestZone then
				local zoneData = currentContinentData[CONTINENT_DATA__ZONES][bestZone]
				local left = zoneData[ZONE_DATA__MAP_LEFT]
				local bottom = zoneData[ZONE_DATA__MAP_BOTTOM]
				local width = zoneData[ZONE_DATA__MAP_WIDTH]
				local height = zoneData[ZONE_DATA__MAP_HEIGHT]
				return bestZone, (x - left) / width, 1 - (y - bottom) / height
			else
				local center = currentContinentData[CONTINENT_DATA__MAP_CENTER]
				local middle = currentContinentData[CONTINENT_DATA__MAP_MIDDLE]
				local width = currentContinentData[CONTINENT_DATA__WIDTH]
				local height = currentContinentData[CONTINENT_DATA__HEIGHT]
				local left = center - width/2
				local bottom = middle - height/2
				return currentContinentData[CONTINENT_DATA__TEXTURE], (x - left) / width, 1 - (y - bottom) / height
			end
		end
	end
	return nil, nil, nil
end
Cartographer.convertUniverseCoordinateToZonePosition = convertUniverseCoordinateToZonePosition

local function convertZonePositionToUniverseCoordinate(zone, x, y)
	if (x == 0 and y == 0) or not x then
		return nil, nil
	end
	local data = ZONE_DATA[zone]
	if data then
		y = 1 - y
		x = x * data[ZONE_DATA__MAP_WIDTH] + data[ZONE_DATA__MAP_LEFT]
		y = y * data[ZONE_DATA__MAP_HEIGHT] + data[ZONE_DATA__MAP_BOTTOM]
		return x, y
	end
	data = CONTINENT_DATA[zone]
	if data then
		y = 1 - y
		local width, height = data[CONTINENT_DATA__WIDTH], data[CONTINENT_DATA__HEIGHT]
		x = x * width + data[CONTINENT_DATA__MAP_CENTER] - width/2
		y = y * height + data[CONTINENT_DATA__MAP_MIDDLE] - height/2
		local zone_z, zone_x, zone_y = convertUniverseCoordinateToZonePosition(x, y)
		local zoneData = ZONE_DATA[zone_z]
		if zoneData then
			return x - zoneData[ZONE_DATA__SHOWN_MAP_LEFT] + zoneData[ZONE_DATA__MAP_LEFT], y - zoneData[ZONE_DATA__SHOWN_MAP_BOTTOM] + zoneData[ZONE_DATA__MAP_BOTTOM]
		end
		return x, y
	end
	return nil, nil
end
Cartographer.convertZonePositionToUniverseCoordinate = convertZonePositionToUniverseCoordinate

local cameraX, cameraY = 0, 0
local cameraZoom = 1
local function getCursorUniverseCoordinate()
	local cursorX, cursorY = GetCursorPosition()
	local mapScale = mapView:GetEffectiveScale()
	cursorX, cursorY = cursorX/mapScale, cursorY/mapScale
	local mapX, mapY, mapW, mapH = mapView:GetRect()
	
	cursorX, cursorY = cursorX - mapX - mapW/2, cursorY - mapY - mapH/2
	return cursorX, cursorY
end
Cartographer.getCursorUniverseCoordinate = getCursorUniverseCoordinate

local last_cx, last_cy
local ret_zz, ret_zx, ret_zy
local function _getCursorZonePosition(cx, cy)
	local zoneData = CURRENT_ZONE_DATA
	if zoneData then
		local left = zoneData[ZONE_DATA__MAP_LEFT]
		local bottom = zoneData[ZONE_DATA__MAP_BOTTOM]
		local width = zoneData[ZONE_DATA__MAP_WIDTH]
		local height = zoneData[ZONE_DATA__MAP_HEIGHT]
		if cx >= left and cy >= bottom and cx <= left+width and cy <= bottom+height then
			local zx, zy = (cx - left) / width, 1 - (cy - bottom) / height
			if not CITIES[CURRENT_MAP_TEXTURE] then
				local name = UpdateMapHighlight(zx, zy)
				local fileName = LOCALIZED_ZONE_TO_TEXTURE[name]
				if fileName then
					zoneData = ZONE_DATA[fileName]
					if zoneData then
						left = zoneData[ZONE_DATA__MAP_LEFT]
						bottom = zoneData[ZONE_DATA__MAP_BOTTOM]
						width = zoneData[ZONE_DATA__MAP_WIDTH]
						height = zoneData[ZONE_DATA__MAP_HEIGHT]
						if cx >= left and cy >= bottom and cx <= left+width and cy <= bottom+height then
							local zx, zy = (cx - left) / width, 1 - (cy - bottom) / height
							return fileName, zx, zy
						end
					end
				end
			end
			return CURRENT_MAP_TEXTURE, zx, zy
		end
	end
	return convertUniverseCoordinateToZonePosition(cx, cy)
end
local function getCursorZonePosition()
	local cx, cy = getCursorUniverseCoordinate()
	if last_cx == cx and last_cy == cy then
		return ret_zz, ret_zx, ret_zy
	end
	ret_zz, ret_zx, ret_zy = _getCursorZonePosition(cx, cy)
	return ret_zz, ret_zx, ret_zy
end
Cartographer.getCursorZonePosition = getCursorZonePosition

local function getUnitUniverseCoordinate(unit)
	local x, y = GetPlayerMapPosition(unit)
	if x == 0 and y == 0 then
		return nil, nil
	end
	return convertZonePositionToUniverseCoordinate(CURRENT_MAP_TEXTURE, x, y)
end
Cartographer.getUnitUniverseCoordinate = getUnitUniverseCoordinate

local function initializeContinentData()
	initializeContinentData = nil
	local continentNames = { GetMapContinents() }
	for i,v in ipairs(continentNames) do
		local data = CONTINENT_DATA[i]
		if not data then
			-- expansion we don't know about
			data = {
				x = 0,
				y = 375,
				width = 225,
				rect = { 0, 1, 0, 1 }
			}
			CONTINENT_DATA[i] = data
		end
		data.name = v
	end
	for continentNum, data in ipairs(CONTINENT_DATA) do
		local currentContinentData = {}
		CONTINENT_DATA[continentNum] = currentContinentData
		SetMapZoom(continentNum)
		currentContinentData[CONTINENT_DATA__TEXTURE] = CURRENT_MAP_TEXTURE
		CONTINENT_DATA[CURRENT_MAP_TEXTURE] = currentContinentData
		currentContinentData[CONTINENT_DATA__INDEX] = continentNum
		
		currentContinentData[CONTINENT_DATA__LOCALIZED_NAME] = data.name
		
		local continentCenter = data.x
		currentContinentData[CONTINENT_DATA__MAP_CENTER] = continentCenter
		
		local continentMiddle = data.y
		currentContinentData[CONTINENT_DATA__MAP_MIDDLE] = continentMiddle
		
		local continentWidth = data.yards / YARDS_PER_PIXEL
		currentContinentData[CONTINENT_DATA__WIDTH] = continentWidth
		
		local continentHeight = continentWidth * BLIZZARD_MAP_HEIGHT_TO_WIDTH_RATIO
		currentContinentData[CONTINENT_DATA__HEIGHT] = continentHeight
		
		local rect = data.rect
		currentContinentData[CONTINENT_DATA__SHOWN_LEFT] = continentCenter + continentWidth * (rect[1] - 0.5)
		currentContinentData[CONTINENT_DATA__SHOWN_RIGHT] = continentCenter + continentWidth * (rect[2] - 0.5)
		currentContinentData[CONTINENT_DATA__SHOWN_TOP] = continentMiddle - continentHeight * (rect[3] - 0.5)
		currentContinentData[CONTINENT_DATA__SHOWN_BOTTOM] = continentMiddle - continentHeight * (rect[4] - 0.5)
		
		local zoneNames = { GetMapZones(continentNum) }
		local zones = {}
		currentContinentData[CONTINENT_DATA__ZONES] = zones
		for num = 1, #zoneNames do
			local x, y
			local name, fileName, texPctX, texPctY, texX, texY, scrollX, scrollY
			
			repeat
				x, y = math.random(), math.random()
				name, fileName, texPctX, texPctY, texX, texY, scrollX, scrollY = UpdateMapHighlight(x, y)
			until fileName and not zones[fileName]
			
			local zones_fileName = {}
			zones_fileName[ZONE_DATA__CONTINENT_NUM] = continentNum
			for i, n in ipairs(zoneNames) do
				if n == name then
					zones_fileName[ZONE_DATA__INDEX] = i
				end
			end
			zones_fileName[ZONE_DATA__LOCALIZED_NAME] = name
			zones_fileName[ZONE_DATA__MAP_WIDTH] = texX
			zones_fileName[ZONE_DATA__MAP_HEIGHT] = texY
			zones_fileName[ZONE_DATA__MAP_LEFT] = scrollX
			zones_fileName[ZONE_DATA__MAP_BOTTOM] = scrollY
			ZONE_DATA[fileName] = zones_fileName
			zones[fileName] = zones_fileName
			LOCALIZED_ZONE_TO_TEXTURE[name] = fileName
			
			for x = scrollX, scrollX+texX, 0.0025 do
				for y = scrollY, scrollY+texY, 0.0025 do
					if UpdateMapHighlight(x, y) == name then
						if not zones_fileName[ZONE_DATA__HIGHLIGHT_LEFT] then
							zones_fileName[ZONE_DATA__HIGHLIGHT_LEFT] = x
							zones_fileName[ZONE_DATA__HIGHLIGHT_RIGHT] = x
							zones_fileName[ZONE_DATA__HIGHLIGHT_BOTTOM] = y
							zones_fileName[ZONE_DATA__HIGHLIGHT_TOP] = y
						else
							if x < zones_fileName[ZONE_DATA__HIGHLIGHT_LEFT] then
								zones_fileName[ZONE_DATA__HIGHLIGHT_LEFT] = x
							elseif x > zones_fileName[ZONE_DATA__HIGHLIGHT_RIGHT] then
								zones_fileName[ZONE_DATA__HIGHLIGHT_RIGHT] = x
							end
							if y < zones_fileName[ZONE_DATA__HIGHLIGHT_BOTTOM] then
								zones_fileName[ZONE_DATA__HIGHLIGHT_BOTTOM] = y
							elseif y > zones_fileName[ZONE_DATA__HIGHLIGHT_TOP] then
								zones_fileName[ZONE_DATA__HIGHLIGHT_TOP] = y
							end
						end
					end
				end
			end
			
			zones_fileName[ZONE_DATA__HIGHLIGHT_LEFT] = zones_fileName[ZONE_DATA__HIGHLIGHT_LEFT]-0.001
			zones_fileName[ZONE_DATA__HIGHLIGHT_RIGHT] = zones_fileName[ZONE_DATA__HIGHLIGHT_RIGHT]+0.001
			zones_fileName[ZONE_DATA__HIGHLIGHT_BOTTOM] = zones_fileName[ZONE_DATA__HIGHLIGHT_BOTTOM]-0.001
			zones_fileName[ZONE_DATA__HIGHLIGHT_TOP] = zones_fileName[ZONE_DATA__HIGHLIGHT_TOP]+0.001
			zones_fileName[ZONE_DATA__SHOWN_MAP_LEFT] = zones_fileName[ZONE_DATA__MAP_LEFT]
			zones_fileName[ZONE_DATA__SHOWN_MAP_BOTTOM] = zones_fileName[ZONE_DATA__MAP_BOTTOM]
			if fileName == "EversongWoods" or fileName == "Ghostlands" or fileName == "Sunwell" or fileName == "SilvermoonCity" then
				zones_fileName[ZONE_DATA__MAP_LEFT] = zones_fileName[ZONE_DATA__MAP_LEFT] - 0.00168
				zones_fileName[ZONE_DATA__MAP_BOTTOM] = zones_fileName[ZONE_DATA__MAP_BOTTOM] + 0.01
				zones_fileName[ZONE_DATA__HIGHLIGHT_LEFT] = zones_fileName[ZONE_DATA__HIGHLIGHT_LEFT] - 0.00168
				zones_fileName[ZONE_DATA__HIGHLIGHT_RIGHT] = zones_fileName[ZONE_DATA__HIGHLIGHT_RIGHT] - 0.00168
				zones_fileName[ZONE_DATA__HIGHLIGHT_BOTTOM] = zones_fileName[ZONE_DATA__HIGHLIGHT_BOTTOM] + 0.01
				zones_fileName[ZONE_DATA__HIGHLIGHT_TOP] = zones_fileName[ZONE_DATA__HIGHLIGHT_TOP] + 0.01
			end
			
			zones_fileName[ZONE_DATA__MAP_BOTTOM] = 1 - zones_fileName[ZONE_DATA__MAP_BOTTOM] - zones_fileName[ZONE_DATA__MAP_HEIGHT]
			zones_fileName[ZONE_DATA__SHOWN_MAP_BOTTOM] = 1 - zones_fileName[ZONE_DATA__SHOWN_MAP_BOTTOM] - zones_fileName[ZONE_DATA__MAP_HEIGHT]
			zones_fileName[ZONE_DATA__MAP_WIDTH] = zones_fileName[ZONE_DATA__MAP_WIDTH] * continentWidth
			zones_fileName[ZONE_DATA__MAP_HEIGHT] = zones_fileName[ZONE_DATA__MAP_HEIGHT] * continentHeight
			zones_fileName[ZONE_DATA__MAP_LEFT] = zones_fileName[ZONE_DATA__MAP_LEFT] * continentWidth - continentWidth/2 + continentCenter
			zones_fileName[ZONE_DATA__MAP_BOTTOM] = zones_fileName[ZONE_DATA__MAP_BOTTOM] * continentHeight - continentHeight/2 + continentMiddle
			zones_fileName[ZONE_DATA__SHOWN_MAP_LEFT] = zones_fileName[ZONE_DATA__SHOWN_MAP_LEFT] * continentWidth - continentWidth/2 + continentCenter
			zones_fileName[ZONE_DATA__SHOWN_MAP_BOTTOM] = zones_fileName[ZONE_DATA__SHOWN_MAP_BOTTOM] * continentHeight - continentHeight/2 + continentMiddle
			zones_fileName[ZONE_DATA__HIGHLIGHT_LEFT] = zones_fileName[ZONE_DATA__HIGHLIGHT_LEFT] * continentWidth - continentWidth/2 + continentCenter
			zones_fileName[ZONE_DATA__HIGHLIGHT_RIGHT] = zones_fileName[ZONE_DATA__HIGHLIGHT_RIGHT] * continentWidth - continentWidth/2 + continentCenter
			zones_fileName[ZONE_DATA__HIGHLIGHT_BOTTOM], zones_fileName[ZONE_DATA__HIGHLIGHT_TOP] = 1 - zones_fileName[ZONE_DATA__HIGHLIGHT_TOP], 1 - zones_fileName[ZONE_DATA__HIGHLIGHT_BOTTOM]
			zones_fileName[ZONE_DATA__HIGHLIGHT_BOTTOM] = zones_fileName[ZONE_DATA__HIGHLIGHT_BOTTOM] * continentHeight - continentHeight/2 + continentMiddle
			zones_fileName[ZONE_DATA__HIGHLIGHT_TOP] = zones_fileName[ZONE_DATA__HIGHLIGHT_TOP] * continentHeight - continentHeight/2 + continentMiddle
		end
		
		currentContinentData[CONTINENT_DATA__MINIMAP_TEXTURE_CENTER_X] = data.minimapTextureCenterX
		currentContinentData[CONTINENT_DATA__MINIMAP_TEXTURE_CENTER_Y] = data.minimapTextureCenterY
		currentContinentData[CONTINENT_DATA__MINIMAP_TEXTURE_OFFSET_X] = data.minimapTextureOffsetX
		currentContinentData[CONTINENT_DATA__MINIMAP_TEXTURE_OFFSET_Y] = data.minimapTextureOffsetY
		currentContinentData[CONTINENT_DATA__MINIMAP_TEXTURE_BACKGROUND] = data.minimapTextureBackground
		local extraMinimapTextureAreas = {}
		currentContinentData[CONTINENT_DATA__EXTRA_MINIMAP_TEXTURE_AREAS] = extraMinimapTextureAreas
		if data.extraMinimapTextureAreas then
			for k,v in pairs(data.extraMinimapTextureAreas) do
				extraMinimapTextureAreas[#extraMinimapTextureAreas+1] = {
					k,
					v.minimapTextureCenterX,
					v.minimapTextureCenterY,
				}
			end
		end
		
		currentContinentData[CONTINENT_DATA__RECT_LEFT] = rect[1]
		currentContinentData[CONTINENT_DATA__RECT_RIGHT] = rect[2]
		currentContinentData[CONTINENT_DATA__RECT_TOP] = rect[3]
		currentContinentData[CONTINENT_DATA__RECT_BOTTOM] = rect[4]
	end
	SetMapToCurrentZone()
	
	local function zoneSearchData_func(alpha, bravo)
		if CITIES[bravo] then
			return false
		elseif CITIES[alpha] then
			return true
		end
		
		local zoneData_alpha = ZONE_DATA[alpha]
		local zoneData_bravo = ZONE_DATA[bravo]
		local area_alpha = (zoneData_alpha[ZONE_DATA__HIGHLIGHT_RIGHT] - zoneData_alpha[ZONE_DATA__HIGHLIGHT_LEFT]) * (zoneData_alpha[ZONE_DATA__HIGHLIGHT_TOP] - zoneData_alpha[ZONE_DATA__HIGHLIGHT_BOTTOM])
		local area_bravo = (zoneData_bravo[ZONE_DATA__HIGHLIGHT_RIGHT] - zoneData_bravo[ZONE_DATA__HIGHLIGHT_LEFT]) * (zoneData_bravo[ZONE_DATA__HIGHLIGHT_TOP] - zoneData_bravo[ZONE_DATA__HIGHLIGHT_BOTTOM])
		return area_alpha < area_bravo
	end
	for continentNum, currentContinentData in ipairs(CONTINENT_DATA) do
		local t = {}
		zoneSearchData[continentNum] = t
		local zones = currentContinentData[CONTINENT_DATA__ZONES]
		
		local continentLeft = currentContinentData[CONTINENT_DATA__SHOWN_LEFT]
		local continentWidth = currentContinentData[CONTINENT_DATA__SHOWN_RIGHT] - currentContinentData[CONTINENT_DATA__SHOWN_LEFT]
		local continentBottom = currentContinentData[CONTINENT_DATA__SHOWN_BOTTOM]
		local continentHeight = currentContinentData[CONTINENT_DATA__SHOWN_TOP] - currentContinentData[CONTINENT_DATA__SHOWN_BOTTOM]
		
		for fileName, zoneData in pairs(zones) do
			local zoneLeft = (zoneData[ZONE_DATA__HIGHLIGHT_LEFT] - continentLeft)/continentWidth
			local zoneRight = (zoneData[ZONE_DATA__HIGHLIGHT_RIGHT] - continentLeft)/continentWidth
			local zoneBottom = (zoneData[ZONE_DATA__HIGHLIGHT_BOTTOM] - continentBottom)/continentHeight
			local zoneTop = (zoneData[ZONE_DATA__HIGHLIGHT_TOP] - continentBottom)/continentHeight
			for x = 0, 9 do
				if zoneLeft < (x+1)/10 and zoneRight > x/10 then
					for y = 0, 9 do
						if zoneBottom < (y+1)/10 and zoneTop > y/10 then
							local num = x*10 + y
							local u = t[num]
							if not u then
								u = {}
								t[num] = u
							end
							u[#u+1] = fileName
						end
					end
				end
			end
		end
		for num = 0, 99 do
			if t[num] then
				table.sort(t[num], zoneSearchData_func)
			end
		end
	end
end

local mapTextures = {}
local function createMapTextures()
	createMapTextures = nil
	if false then
		for i = 1, 12 do
			local tex = mapView:CreateTexture(mapView:GetName() .. "_MapTexture_Cosmic_" .. i, "BACKGROUND")
			mapTextures[#mapTextures+1] = tex
			tex:SetTexture(([=[Interface\WorldMap\Cosmic\cosmic%d]=]):format(i))
			local continentTileSize = 120
			local continentWidth = 480
			local continentHeight = 360
			local x = ((i-1) % 4) * continentTileSize - continentWidth/2
			local y = -math.floor((i-1) / 4 + 1) * continentTileSize + continentHeight/2
			local x1 = x
			local x2 = x1 + continentTileSize
			local y1 = y
			local y2 = y1 + continentTileSize
			tex:SetPoint("BOTTOMLEFT", mapView, "CENTER", x1, y1)
			tex:SetPoint("TOPRIGHT", mapView, "CENTER", x2, y2)
		end
	end
	for continentNum, currentContinentData in ipairs(CONTINENT_DATA) do
		local continentCenter = currentContinentData[CONTINENT_DATA__MAP_CENTER]
		local continentMiddle = currentContinentData[CONTINENT_DATA__MAP_MIDDLE]
		local continentWidth = currentContinentData[CONTINENT_DATA__WIDTH]
		local continentHeight = currentContinentData[CONTINENT_DATA__HEIGHT]
		local continentTileSize = continentWidth * 256/BLIZZARD_MAP_WIDTH
		local rectLeft = currentContinentData[CONTINENT_DATA__RECT_LEFT]
		local rectRight = currentContinentData[CONTINENT_DATA__RECT_RIGHT]
		local rectTop = currentContinentData[CONTINENT_DATA__RECT_TOP]
		local rectBottom = currentContinentData[CONTINENT_DATA__RECT_BOTTOM]
		for i = 1, 12 do
			local good = true
			local left = currentContinentData[CONTINENT_DATA__RECT_LEFT]*1002/1024 * 4 - ((i-1)%4)
			if left < 0 then
				left = 0
			elseif left >= 1 then
				good = false
			end
			local right = currentContinentData[CONTINENT_DATA__RECT_RIGHT]*1002/1024 * 4 - ((i-1)%4)
			if right > 1 then
				right = 1
			elseif right <= 0 then
				good = false
			end
			local top = currentContinentData[CONTINENT_DATA__RECT_TOP]*668/768 * 3 - math.floor((i-1) / 4)
			if top < 0 then
				top = 0
			elseif top >= 1 then
				good = false
			end
			local bottom =  currentContinentData[CONTINENT_DATA__RECT_BOTTOM]*668/768 * 3 - math.floor((i-1) / 4)
			if bottom > 1 then
				bottom = 1
			elseif bottom <= 0 then
				good = false
			end
			
			if good then
				local x = ((i-1) % 4) * continentTileSize + continentCenter - continentWidth/2
				local y = -math.floor((i-1) / 4 + 1) * continentTileSize + continentMiddle + continentHeight/2
				local x1 = x + left * continentTileSize
				local x2 = x + right * continentTileSize
				local y1 = y + (1 - bottom) * continentTileSize
				local y2 = y + (1 - top) * continentTileSize
				
				local tex = mapView:CreateTexture(mapView:GetName() .. "_MapTexture_" .. currentContinentData[CONTINENT_DATA__TEXTURE] .. i, "BORDER")
				mapTextures[#mapTextures+1] = tex
				local mapFileName = currentContinentData[CONTINENT_DATA__TEXTURE]
				tex:SetTexture(([=[Interface\WorldMap\%s\%s%d]=]):format(mapFileName, mapFileName, i))
				tex:SetPoint("BOTTOMLEFT", mapView, "CENTER", x1, y1)
				tex:SetPoint("TOPRIGHT", mapView, "CENTER", x2, y2)
			
				tex:SetTexCoord(left, right, top, bottom)
			end
		end
		
		-- no longer use rect data
		currentContinentData[CONTINENT_DATA__RECT_LEFT] = nil
		currentContinentData[CONTINENT_DATA__RECT_RIGHT] = nil
		currentContinentData[CONTINENT_DATA__RECT_TOP] = nil
		currentContinentData[CONTINENT_DATA__RECT_BOTTOM] = nil
	end
end

local cos, sin, abs = cos, sin, math.abs
local function smallRotate(angle, x, y)
	local A = cos(angle)
	local B = sin(angle)
	return x * A - y * B, x * B + y * A
end

local function rotate(angle)
	local A = cos(angle)
	local B = sin(angle)
	local ULx, ULy = -0.5 * A - -0.5 * B, -0.5 * B + -0.5 * A
	local LLx, LLy = -0.5 * A - 0.5 * B, -0.5 * B + 0.5 * A
	local URx, URy = 0.5 * A - -0.5 * B, 0.5 * B + -0.5 * A
	local LRx, LRy = 0.5 * A - 0.5 * B, 0.5 * B + 0.5 * A
	return ULx+0.5, ULy+0.5, LLx+0.5, LLy+0.5, URx+0.5, URy+0.5, LRx+0.5, LRy+0.5
end

local shownPOIs = {}
local function OnShow(this)
	shownPOIs[this] = true
end
local function OnHide(this)
	shownPOIs[this] = nil
end
local tmp = {}
local function poi_sort(alpha, bravo)
	return alpha:GetName() < bravo:GetName()
end
local inPOI = nil
local function OnEnter(this)
	inPOI = 0
end
local function OnLeave(this)
	inPOI = 0
end
local alreadySetOwner = false
function Cartographer.UpdatePOITooltips(elapsed, currentTime)
	if not inPOI or inPOI > currentTime then
		return
	end
	inPOI = currentTime + 0.5
	if not alreadySetOwner then
		WorldMapTooltip:SetOwner(mapHolder, "ANCHOR_CURSOR")
		alreadySetOwner = true
	end
	for poi in pairs(shownPOIs) do
		if IsMouseHovering(poi) then
			tmp[#tmp+1] = poi
		end
	end
	WorldMapTooltip:ClearLines()
	if #tmp == 0 then
		inPOI = nil
		alreadySetOwner = false
		WorldMapTooltip:Hide()
	elseif #tmp == 1 then
		tmp[1]:AddDataToFullTooltip()
		tmp[1] = nil
	else
		table.sort(tmp, poi_sort)
		for i, v in ipairs(tmp) do
			v:AddDataToTooltipLine()
			tmp[i] = nil
		end
	end
	WorldMapTooltip:Show()
end
local totalPOIs = setmetatable({}, {__newindex=function(self, key, value)
	assert(value == true)
	if key:IsShown() then
		shownPOIs[key] = true
	end
	key:SetScript("OnShow", OnShow)
	key:SetScript("OnHide", OnHide)
	key:SetScript("OnEnter", OnEnter)
	key:SetScript("OnLeave", OnLeave)
	key:EnableMouse(true)
end})
local playerArrow
local groupPOIs = {}
local guildPOIs = {}
local corpsePOI, rezPOI

local function readjustCamera()
	local previousScale = mapView:GetScale()
	if previousScale ~= cameraZoom then
		mapView:SetScale(cameraZoom)
		
		local size = 20 / cameraZoom
		playerArrow:SetWidth(size * 2^0.5)
		playerArrow:SetHeight(size * 2^0.5)
		if corpsePOI then
			corpsePOI:SetWidth(size)
			corpsePOI:SetHeight(size)
			rezPOI:SetWidth(size)
			rezPOI:SetHeight(size)
		end
		for i, poi in ipairs(groupPOIs) do
			poi:SetWidth(size)
			poi:SetHeight(size)
		end
		for name, poi in pairs(guildPOIs) do
			poi:SetWidth(size)
			poi:SetHeight(size)
		end
	end
	scrollFrame:SetHorizontalScroll(-(mapView:GetWidth() - scrollFrame:GetWidth()/cameraZoom) / 2 - cameraX)
	scrollFrame:SetVerticalScroll((mapView:GetHeight() - scrollFrame:GetHeight()/cameraZoom) / 2 - cameraY)
end
Cartographer.readjustCamera = readjustCamera

local function clampMapCamera()
	do return end
	local universeLeft, universeRight, universeTop, universeBottom = 0, 0, 0, 0
	for continentNum, currentContinentData in ipairs(CONTINENT_DATA) do
		local continentLeft = currentContinentData[CONTINENT_DATA__SHOWN_LEFT]
		if continentLeft < universeLeft then
			universeLeft = continentLeft
		end
		
		local continentRight = currentContinentData[CONTINENT_DATA__SHOWN_RIGHT]
		if continentRight > universeRight then
			universeRight = continentRight
		end
		
		local continentBottom = currentContinentData[CONTINENT_DATA__SHOWN_BOTTOM]
		if continentBottom < universeBottom then
			universeBottom = continentBottom
		end
		
		local continentTop = currentContinentData[CONTINENT_DATA__SHOWN_TOP]
		if continentTop > universeTop then
			universeTop = continentTop
		end
	end
	
	local scrollFrame_width = scrollFrame:GetWidth()
	local scrollFrame_height = scrollFrame:GetHeight()
	
	local mapViewLeft = cameraX - scrollFrame_width/cameraZoom/2
	local mapViewRight = cameraX + scrollFrame_width/cameraZoom/2
	local mapViewBottom = cameraY - scrollFrame_height/cameraZoom/2
	local mapViewTop = cameraY + scrollFrame_height/cameraZoom/2
	
	if universeRight - universeLeft < mapViewRight - mapViewLeft then
		if universeLeft < mapViewLeft then
			cameraX = cameraX + universeLeft - mapViewLeft
		elseif universeRight > mapViewRight then
			cameraX = cameraX + universeRight - mapViewRight
		end
	else
		if universeLeft > mapViewLeft then
			if universeRight > mapViewRight then
				cameraX = cameraX + universeLeft - mapViewLeft
			end
		elseif universeRight < mapViewRight then
			cameraX = cameraX + universeRight - mapViewRight
		end
	end
	
	if universeTop - universeBottom < mapViewTop - mapViewBottom then
		if universeBottom < mapViewBottom then
			cameraY = cameraY + universeBottom - mapViewBottom
		elseif universeTop > mapViewTop then
			cameraY = cameraY + universeTop - mapViewTop
		end
	else
		if universeBottom > mapViewBottom then
			if universeTop > mapViewTop then
				cameraY = cameraY + universeBottom - mapViewBottom
			end
		elseif universeTop < mapViewTop then
			cameraY = cameraY + universeTop - mapViewTop
		end
	end
end

local instanceTextureFrames = {}
local loadInstanceTextureSection
do
	local doneInstances = {}
	local workingInstance, workingInstancePosition
	local first = true
	
	local function unpack_float(D, C, B, A)
		local negative = A >= 128 and -1 or 1
		local exponent = 2*(A%128) + math.floor(B/128) - 127
		local mantissa = 1 + (65536*(B%128) + 256*C + D) / 2^23
		return negative * 2^exponent * mantissa
	end
	
	function loadInstanceTextureSection()
		if first then
			first = false
			for name, data in pairs(INSTANCE_TEXTURE_DATA) do
				instanceTextureFrame = CreateFrame("Frame", mapView:GetName() .. "_InstanceTextureFrame_" .. name:gsub("[ ']", ""), mapView)
				instanceTextureFrames[name] = instanceTextureFrame
				local size = 0.01
				instanceTextureFrame:SetScale(0.01)
				local offsetX = (200 + math.random()*400) / size-- -currentContinentData[CONTINENT_DATA__MINIMAP_TEXTURE_OFFSET_X] + currentContinentData[CONTINENT_DATA__MAP_CENTER]
				local offsetY = (-200 + math.random()*400) / size-- currentContinentData[CONTINENT_DATA__MINIMAP_TEXTURE_OFFSET_Y] + currentContinentData[CONTINENT_DATA__MAP_MIDDLE]
				instanceTextureFrame:SetPoint("CENTER", mapView, "CENTER", offsetX, offsetY)
				instanceTextureFrame:SetWidth(1e-10)
				instanceTextureFrame:SetHeight(1e-10)
				if false then
					-- TODO: fill instance with black background
					local tex = minimapStyleTextureFrame:CreateTexture(minimapStyleTextureFrame:GetName() .. "_Background", "BACKGROUND")
					tex:SetTexture(([=[textures\Minimap\%s]=]):format(currentContinentData[CONTINENT_DATA__MINIMAP_TEXTURE_BACKGROUND]))
					tex:SetPoint("BOTTOMLEFT", minimapStyleTextureFrame, "CENTER", currentContinentData[CONTINENT_DATA__SHOWN_LEFT] - offsetX, currentContinentData[CONTINENT_DATA__SHOWN_BOTTOM] - offsetY)
					tex:SetPoint("TOPRIGHT", minimapStyleTextureFrame, "CENTER", currentContinentData[CONTINENT_DATA__SHOWN_RIGHT] - offsetX, currentContinentData[CONTINENT_DATA__SHOWN_TOP] - offsetY)
				end
				instanceTextureFrame:Hide()
			end
			return
		end	
		if not workingInstance then
			for name, data in pairs(INSTANCE_TEXTURE_DATA) do
				if not doneInstances[name] then
					workingInstance = name
				end
			end
			if not workingInstance then
				-- all done
				removeTimer(loadInstanceTextureSection)
				loadInstanceTextureSection = nil
				return
			end
			workingInstancePosition = 0
		end
		
		local workingInstanceData = INSTANCE_TEXTURE_DATA[workingInstance]
		local instanceTextureFrame = instanceTextureFrames[workingInstance]
		
		local angle = INSTANCE_ROTATIONS[workingInstance] or 0
		angle = angle - 90
		local ULx, ULy, LLx, LLy, URx, URy, LRx, LRy = rotate(angle)
		
		local finish = GetTime() + ITERATIVE_PROCESS_STEP_SECONDS
		while GetTime() < finish do
			local offset = workingInstancePosition * 28
			if offset >= #workingInstanceData then
				workingInstancePosition = nil
				break
			end
			workingInstancePosition = workingInstancePosition + 1
			
			local texturePath = ("%02x%02x%02x%02x%02x%02x%02x%02x%02x%02x%02x%02x%02x%02x%02x%02x"):format(workingInstanceData:byte(offset + 1, offset + 16))
			
			local x = unpack_float(workingInstanceData:byte(offset + 17, offset + 20))
			local y = unpack_float(workingInstanceData:byte(offset + 21, offset + 24))
			local z = unpack_float(workingInstanceData:byte(offset + 25, offset + 28))
			
			local tex = instanceTextureFrame:CreateTexture(minimapStyleTextureFrame:GetName() .. "_Texture" .. x .. "_" .. y .. "_" .. z, "ARTWORK")
			instanceTextureFrame[#instanceTextureFrame+1] = tex
			--tex:Hide()
			tex:SetTexture(([=[textures\Minimap\%s]=]):format(texturePath))
			tex:SetAlpha(1)
			tex:SetPoint("CENTER", instanceTextureFrame, "CENTER", smallRotate(angle, x, y))
			tex:SetTexCoord(ULx, ULy, LLx, LLy, URx, URy, LRx, LRy)
		end
		if not workingInstancePosition then
			INSTANCE_TEXTURE_DATA[workingInstance] = nil
			doneInstances[workingInstance] = true
			workingInstance = nil
		end
	end
end

local minimapStyleTextureFrames = {}
local loadMinimapStyleTextureSection
do
	local doneContinents = {}
	local workingContinent, workingExtraArea, workingTextureDataPosition
	local first = true
	function loadMinimapStyleTextureSection()
		if first then
			first = false
			for i, currentContinentData in ipairs(CONTINENT_DATA) do
				local currentContinentData_texture = currentContinentData[CONTINENT_DATA__TEXTURE]
				minimapStyleTextureFrame = CreateFrame("Frame", mapView:GetName() .. "_MinimapStyleTextureFrame_" .. currentContinentData_texture, mapView)
				minimapStyleTextureFrames[currentContinentData_texture] = minimapStyleTextureFrame
				local offsetX = -currentContinentData[CONTINENT_DATA__MINIMAP_TEXTURE_OFFSET_X] + currentContinentData[CONTINENT_DATA__MAP_CENTER]
				local offsetY = currentContinentData[CONTINENT_DATA__MINIMAP_TEXTURE_OFFSET_Y] + currentContinentData[CONTINENT_DATA__MAP_MIDDLE]
				minimapStyleTextureFrame:SetPoint("CENTER", mapView, "CENTER", offsetX, offsetY)
				minimapStyleTextureFrame:SetWidth(1e-10)
				minimapStyleTextureFrame:SetHeight(1e-10)
				if currentContinentData[CONTINENT_DATA__MINIMAP_TEXTURE_BACKGROUND] then
					local tex = minimapStyleTextureFrame:CreateTexture(minimapStyleTextureFrame:GetName() .. "_Background", "BACKGROUND")
					tex:SetTexture(([=[textures\Minimap\%s]=]):format(currentContinentData[CONTINENT_DATA__MINIMAP_TEXTURE_BACKGROUND]))
					tex:SetPoint("BOTTOMLEFT", minimapStyleTextureFrame, "CENTER", currentContinentData[CONTINENT_DATA__SHOWN_LEFT] - offsetX, currentContinentData[CONTINENT_DATA__SHOWN_BOTTOM] - offsetY)
					tex:SetPoint("TOPRIGHT", minimapStyleTextureFrame, "CENTER", currentContinentData[CONTINENT_DATA__SHOWN_RIGHT] - offsetX, currentContinentData[CONTINENT_DATA__SHOWN_TOP] - offsetY)
				end
				minimapStyleTextureFrame:Hide()
				
			end
			return
		end
		if not workingContinent then
			if CONTINENT_DATA[CURRENT_MAP_TEXTURE] then
				workingContinent = CONTINENT_DATA[CURRENT_MAP_TEXTURE][CONTINENT_DATA__INDEX]
			elseif CURRENT_ZONE_DATA then
				workingContinent = CURRENT_ZONE_DATA[ZONE_DATA__CONTINENT_NUM]
			end
			
			local isDone
			if workingContinent then
				local continentTexture = CONTINENT_DATA[workingContinent][CONTINENT_DATA__TEXTURE]
				isDone = doneContinents[continentTexture]
			end
			if isDone or not workingContinent then
				workingContinent = nil
				for continentNum, currentContinentData in ipairs(CONTINENT_DATA) do
					if not doneContinents[currentContinentData[CONTINENT_DATA__TEXTURE]] then
						workingContinent = continentNum
					end
				end
			end
			if not workingContinent then
				-- all done
				removeTimer(loadMinimapStyleTextureSection)
				loadMinimapStyleTextureSection = nil
				return
			end
		end
	
		local currentContinentData = CONTINENT_DATA[workingContinent]
		local currentContinentData_texture = currentContinentData[CONTINENT_DATA__TEXTURE]
		local minimapStyleTextureFrame = minimapStyleTextureFrames[currentContinentData_texture]
		
		local size = BLIZZARD_MINIMAP_TILE_YARD_SIZE / YARDS_PER_PIXEL
		local centerX, centerY
		
		local currentTextureData
		if not workingExtraArea then
			currentTextureData = TEXTURE_DATA[currentContinentData_texture]
			centerX, centerY = currentContinentData[CONTINENT_DATA__MINIMAP_TEXTURE_CENTER_X], currentContinentData[CONTINENT_DATA__MINIMAP_TEXTURE_CENTER_Y]
		else
			local extraMinimapTextureArea = currentContinentData[CONTINENT_DATA__EXTRA_MINIMAP_TEXTURE_AREAS][workingExtraArea]
			currentTextureData = TEXTURE_DATA[extraMinimapTextureArea[1]]
			centerX, centerY = extraMinimapTextureArea[2], extraMinimapTextureArea[3]
		end
	
		local finish = GetTime() + ITERATIVE_PROCESS_STEP_SECONDS
		while GetTime() < finish do
			local texturePath
			workingTextureDataPosition, texturePath = next(currentTextureData, workingTextureDataPosition)
			if not workingTextureDataPosition then
				break
			end
			local x, y = math.floor(workingTextureDataPosition / 100), workingTextureDataPosition % 100
			local chunk_num = math.floor(x - centerX) * 100 + math.floor(y - centerY) + 10000
			local tex = minimapStyleTextureFrame:CreateTexture(minimapStyleTextureFrame:GetName() .. "_Texture" .. chunk_num, "ARTWORK")
			minimapStyleTextureFrame[chunk_num] = tex
			tex:Hide()
			tex:SetTexture(([=[textures\Minimap\%s]=]):format(texturePath))
			tex:SetWidth(size)
			tex:SetHeight(size)
			tex:SetAlpha(1)
			tex:SetPoint("BOTTOMLEFT", minimapStyleTextureFrame, "CENTER", (x - centerX) * size, -(y - centerY) * size)
		end
		if not workingTextureDataPosition then
			TEXTURE_DATA[currentContinentData[CONTINENT_DATA__TEXTURE]] = nil
			doneContinents[currentContinentData[CONTINENT_DATA__TEXTURE]] = true
			if not workingExtraArea then
				workingExtraArea = 1
			else
				workingExtraArea = workingExtraArea + 1
			end
			
			if not currentContinentData[CONTINENT_DATA__EXTRA_MINIMAP_TEXTURE_AREAS][workingExtraArea] then
				-- done continent
				workingContinent = nil
				workingExtraArea = nil
				-- no longer using minimap texture data, clear it, refresh the table
				currentContinentData[CONTINENT_DATA__MINIMAP_TEXTURE_CENTER_X] = nil
				currentContinentData[CONTINENT_DATA__MINIMAP_TEXTURE_CENTER_Y] = nil
				currentContinentData[CONTINENT_DATA__MINIMAP_TEXTURE_OFFSET_X] = nil
				currentContinentData[CONTINENT_DATA__MINIMAP_TEXTURE_OFFSET_Y] = nil
				currentContinentData[CONTINENT_DATA__MINIMAP_TEXTURE_BACKGROUND] = nil
				currentContinentData[CONTINENT_DATA__EXTRA_MINIMAP_TEXTURE_AREAS] = nil
				local newData = { unpack(currentContinentData) } -- this is done because the old table can't get rid of its old size, since we're finally stable, this'll be the new data.
				for k, v in pairs(CONTINENT_DATA) do
					if v == currentContinentData then
						CONTINENT_DATA[k] = newData
					end
				end
			end
		end
	end
end

local zoneFrames = {}
local loadMapOverlayTextureSection
do
	local workingZone, workingTName
	local doneZones = {}
	local first = true
	function loadMapOverlayTextureSection()
		if first then
			first = false
			
			local order = {}
			for zoneTexture in pairs(ZONE_DATA) do
				order[#order+1] = zoneTexture
			end
			table.sort(order, function(alpha, bravo)
				if CITIES[alpha] then
					return false
				elseif CITIES[bravo] then
					return true
				else
					return alpha > bravo
				end
			end)
			
			for _, zoneTexture in ipairs(order) do
				local zoneData = ZONE_DATA[zoneTexture]
				local zoneFrame = CreateFrame("Frame", mapView:GetName() .. "_ZoneFrame_" .. zoneTexture, mapView)
				zoneFrame.texture = zoneTexture
				zoneFrame:Hide()
				zoneFrame:SetAlpha(0)
				zoneFrame:SetWidth(BLIZZARD_MAP_WIDTH)
				zoneFrame:SetHeight(BLIZZARD_MAP_WIDTH * BLIZZARD_MAP_HEIGHT_TO_WIDTH_RATIO)
				zoneFrames[zoneTexture] = zoneFrame
				local scale = zoneData[ZONE_DATA__MAP_WIDTH] / BLIZZARD_MAP_WIDTH
				zoneFrame:SetScale(scale)
				zoneFrame:SetPoint("BOTTOMLEFT", mapView, "CENTER", zoneData[ZONE_DATA__MAP_LEFT] / scale, zoneData[ZONE_DATA__MAP_BOTTOM] / scale)
			end
			return
		end
		if not workingZone then
			if CURRENT_ZONE_DATA then
				workingZone = CURRENT_MAP_TEXTURE
			end
			
			if not workingZone or doneZones[workingZone] then
				workingZone = nil
				for zoneTexture in pairs(ZONE_DATA) do
					if not doneZones[zoneTexture] then
						workingZone = zoneTexture
						break
					end
				end
				if not workingZone then
					-- all done
					removeTimer(loadMapOverlayTextureSection)
					loadMapOverlayTextureSection = nil
					return
				end
			end
		end
		local zoneFrame = zoneFrames[workingZone]
		assert(zoneFrame)
		local overlayData = OVERLAY_DATA[workingZone]
		doneZones[workingZone] = true
		
		if overlayData then
			local finishTime = GetTime() + ITERATIVE_PROCESS_STEP_SECONDS
			while GetTime() < finishTime do
				local num
				workingTName, num = next(overlayData, workingTName)
				if not workingTName then
					break
				end
				
				local textureName = ([=[Interface\WorldMap\%s\%s]=]):format(workingZone, workingTName)
				local textureWidth = num % 1024
				local textureHeight = math.floor(num / 1024) % 1024
				local offsetX = math.floor(num / (1024*1024)) % 1024
				local offsetY = math.floor(num / (1024*1024*1024)) % 1024

				if textureName == [=[Interface\WorldMap\Tirisfal\BRIGHTWATERLAKE]=] then
					if offsetX == 587 then
						offsetX = 584
					end
				elseif textureName == [=[Interface\WorldMap\Silverpine\BERENSPERIL]=] then
					if offsetY == 417 then
						offsetY = 415
					end
				end

				local numTexturesWide = math.ceil(textureWidth / 256)
				local numTexturesTall = math.ceil(textureHeight / 256)

				for j = 1, numTexturesTall do
					local texturePixelHeight
					local textureFileHeight
					if j < numTexturesTall then
						texturePixelHeight = 256
						textureFileHeight = 256
					else
						texturePixelHeight = textureHeight % 256
						if texturePixelHeight == 0 then
							texturePixelHeight = 256
						end
						textureFileHeight = 16
						while textureFileHeight < texturePixelHeight do
							textureFileHeight = textureFileHeight * 2
						end
					end
					for k = 1, numTexturesWide do
						local texturePixelWidth
						local textureFileWidth
						if k < numTexturesWide then
							texturePixelWidth = 256
							textureFileWidth = 256
						else
							texturePixelWidth = textureWidth % 256
							if texturePixelWidth == 0 then
								texturePixelWidth = 256
							end
							textureFileWidth = 16
							while textureFileWidth < texturePixelWidth do
								textureFileWidth = textureFileWidth * 2
							end
						end

						local texture = zoneFrame:CreateTexture(zoneFrame:GetName() .. "_Texture" .. (#zoneFrame+1), "ARTWORK")
						zoneFrame[#zoneFrame+1] = texture
						texture:SetWidth(texturePixelWidth)
						texture:SetHeight(texturePixelHeight)
						texture:SetTexCoord(0, texturePixelWidth/textureFileWidth, 0, texturePixelHeight/textureFileHeight)
						texture:SetPoint("TOPLEFT", zoneFrame, "TOPLEFT", (offsetX + (256 * (k-1))), -(offsetY + (256 * (j - 1))))
						texture:SetTexture(textureName..(((j - 1) * numTexturesWide) + k))

						-- if discovered[tname] then
						-- 	texture:SetVertexColor(1.0,1.0,1.0)
						-- 	texture:SetAlpha(1.0)
						-- else
						-- 	texture:SetVertexColor(self.db.profile.darkR, self.db.profile.darkG, self.db.profile.darkB)
						-- 	texture:SetAlpha(self.db.profile.darkA)
						-- end
						texture:Show()
					end
				end
			end
			if not workingTName then
				workingZone = nil
			end
		elseif CITIES[workingZone] then
			for i = 1, 12 do
				local texture = zoneFrame:CreateTexture(zoneFrame:GetName() .. "_Texture" .. (#zoneFrame+1), "ARTWORK")
				zoneFrame[#zoneFrame+1] = texture
				texture:SetWidth(256)
				texture:SetHeight(256)
				texture:SetTexture(([=[Interface\WorldMap\%s\%s%d]=]):format(workingZone, workingZone, i))
				texture:SetPoint("TOPLEFT", zoneFrame, "TOPLEFT", ((i - 1) % 4) * 256, -math.floor((i - 1) / 4) * 256)
			end
			workingZone = nil
		else
			workingZone = nil
		end
	end
end

local positionFontString
local timeToUpdate = 0
function Cartographer.UpdatePositionFontString(elapsed, currentTime)
	timeToUpdate = timeToUpdate - elapsed
	if timeToUpdate > 0 then
		return
	else
		timeToUpdate = 0.1
	end
	local cz, cx, cy
	if IsMouseHovering(scrollFrame) then
		cz, cx, cy = getCursorZonePosition()
		if not cx then
			cx, cy = 0, 0
		end
	else
		cz, cx, cy = CURRENT_MAP_TEXTURE, GetPlayerMapPosition("player")
	end
	
	local zoneData = ZONE_DATA[cz]
	local zoneName
	if zoneData then
		zoneName = zoneData[ZONE_DATA__LOCALIZED_NAME]
	else
		local continentData = CONTINENT_DATA[cz]
		if continentData then
			zoneName = continentData[CONTINENT_DATA__LOCALIZED_NAME]
		end
	end
	local subzone
	if CURRENT_MAP_TEXTURE == cz and (cx ~= 0 or cy ~= 0) then
		subzone = UpdateMapHighlight(cx, cy)
	end
	local formatString
	if zoneName then
		if cx ~= 0 or cy ~= 0 then
			if subzone then
				formatString = "%1$s: %2$s (%3$.1f%5$s%4$.1f)"
			else
				formatString = "%1$s (%3$.1f%5$s%4$.1f)"
			end
		else
			if subzone then
				formatString = "%1$s: %2$s"
			else
				formatString = "%1$s"
			end
		end
	else
		formatString = ""
	end
	if formatString == "" then
		positionFontString:SetText("")
	else
		positionFontString:SetFormattedText(formatString, zoneName, subzone, cx*100, cy*100, COORDINATE_SEPARATOR)
	end
end

local moving, mouseDown = false, false
function Cartographer.ShowUIOnHover(elapsed, currentTime)
	local alpha = mapFrame:GetAlpha()
	if moving or mouseDown or IsMouseHovering(mapFrame) then
		if alpha == 1 then
			return
		end
		alpha = alpha + elapsed/FADE_TIME
		if alpha > 1 then
			alpha = 1
		end
	else
		if alpha == 0 then
			return
		end
		alpha = alpha - elapsed/FADE_TIME
		if alpha < 0 then
			alpha = 0
		end
	end
	mapFrame:SetAlpha(alpha)
end

function Cartographer.FixViewIfMoving(elapsed, currentTime)
	if moving then
		clampMapCamera()
		readjustCamera()
	end
end

local sizingPoint
function Cartographer.ChangeCursorIfOnSide(elapsed, currentTime)
	if not mouseDown and mapHolder == GetMouseFocus() then
		local cx, cy = getScaledCursorPosition()
		local x, y, w, h = mapHolder:GetRect()
		local bottomSize = cy >= y and cy <= y + 10
		local topSize = cy <= y + h and cy >= y + h - 10
		local leftSize = cx >= x and cx <= x + 10
		local rightSize = cx <= x + w and cx >= x + w - 10
		
		sizingPoint = nil
		if topSize then
			sizingPoint = "TOP"
		elseif bottomSize then
			sizingPoint = "BOTTOM"
		end
		if leftSize then
			sizingPoint = (sizingPoint or '') .. "LEFT"
		elseif rightSize then
			sizingPoint = (sizingPoint or '') .. "RIGHT"
		end
		if sizingPoint then
			SetCursor([=[Interface\Cursor\Inspect]=])
		else
			SetCursor("POINT_CURSOR")
		end
	end
end

local math_cos, math_sin, math_pi = math.cos, math.sin, math.pi

local lastAngle
local inCombat = false
local isDead = not not UnitIsDeadOrGhost("player")
local flashing = true
local nextPlayerArrowColorUpdate = 0
function Cartographer.MovePlayerArrow(elapsed, currentTime)
	local playerX, playerY = getUnitUniverseCoordinate("player")
	
	if playerX then
		playerArrow:SetPoint("CENTER", mapView, "CENTER", playerX, playerY)
	end
	local angle = GetPlayerRotation()
	if angle ~= lastAngle then
		lastAngle = angle
		local x_1 = math_cos(angle + math_pi*3/4)
		local y_1 = math_sin(angle + math_pi*3/4)
		local alpha = x_1 + 0.5
		local bravo = -y_1 + 0.5
		local charlie = -x_1 + 0.5
		local delta = y_1 + 0.5
		playerArrow.inner:SetTexCoord(
			alpha, bravo,
			bravo, charlie,
			delta, alpha,
			charlie, delta
		)
		playerArrow.outer:SetTexCoord(
			alpha, bravo,
			bravo, charlie,
			delta, alpha,
			charlie, delta
		)
	end
	if currentTime > nextPlayerArrowColorUpdate then
		nextPlayerArrowColorUpdate = currentTime + 0.5
		flashing = not flashing
		if flashing then
			if inCombat then
				playerArrow.outer:SetVertexColor(unpack(COMBAT_COLOR))
			elseif isDead then
				playerArrow.outer:SetVertexColor(unpack(DEAD_COLOR))
			else
				playerArrow.outer:SetVertexColor(unpack(NORMAL_STATUS_COLOR))
			end
		else
			playerArrow.outer:SetVertexColor(unpack(NORMAL_STATUS_COLOR))
		end
		playerArrow:SetParent(nil)
		playerArrow:SetParent(mapView)
	end
end

function Cartographer:PLAYER_REGEN_ENABLED()
	inCombat = false
end

function Cartographer:PLAYER_REGEN_DISABLED()
	inCombat = true
end

function Cartographer:PLAYER_DEAD()
	isDead = not not UnitIsDeadOrGhost("player")
	if isDead then
		if not corpsePOI then
			corpsePOI = CreateFrame("Frame", mapView:GetName() .. "_Corpse", mapView)
			totalPOIs[corpsePOI] = true
			function corpsePOI:AddDataToFullTooltip()
				WorldMapTooltip:SetText(CORPSE_RED)
			end
			corpsePOI.AddDataToTooltipLine = corpsePOI.AddDataToFullTooltip
			corpsePOI:SetWidth(1)
			corpsePOI:SetHeight(1)
			local texture = corpsePOI:CreateTexture(corpsePOI:GetName() .. "_Texture", "BORDER")
			corpsePOI.texture = texture
			texture:SetAllPoints()
			texture:SetTexture([[Interface\Minimap\POIIcons]])
			texture:SetTexCoord(0.875, 1, 0, 0.125)
			
			rezPOI = CreateFrame("Frame", mapView:GetName() .. "_DeathRelease", mapView)
			function rezPOI:AddDataToFullTooltip()
				WorldMapTooltip:SetText(SPIRIT_HEALER_RELEASE_RED)
			end
			rezPOI.AddDataToTooltipLine = rezPOI.AddDataToFullTooltip
			totalPOIs[rezPOI] = true
			rezPOI:SetWidth(1)
			rezPOI:SetHeight(1)
			local texture = rezPOI:CreateTexture(rezPOI:GetName() .. "_Texture", "BORDER")
			rezPOI.texture = texture
			texture:SetAllPoints()
			texture:SetTexture([[Interface\Minimap\POIIcons]])
			texture:SetTexCoord(0.875, 1, 0, 0.125)
		end
		rezPOI:Show()
		
		local corpseX, corpseY = convertZonePositionToUniverseCoordinate(CURRENT_MAP_TEXTURE, GetCorpseMapPosition())
		if not corpseX then
			corpsePOI:Hide()
		else
			corpsePOI:SetPoint("CENTER", mapView, "CENTER", corpseX, corpseY)
			corpsePOI:Show()
		end
		
		local rezX, rezY = convertZonePositionToUniverseCoordinate(CURRENT_MAP_TEXTURE, GetDeathReleasePosition())
		if not rezX then
			rezPOI:Hide()
		else
			rezPOI:SetPoint("CENTER", mapView, "CENTER", rezX, rezY)
			rezPOI:Show()
		end
	else
		if corpsePOI then
			corpsePOI:Hide()
			rezPOI:Hide()
		end
	end
end
Cartographer.PLAYER_ALIVE = Cartographer.PLAYER_DEAD
Cartographer.PLAYER_UNGHOST = Cartographer.PLAYER_DEAD

local dragStartX, dragStartY, dragging
function Cartographer.DragMap(elapsed, currentTime)
	if dragging then
		local cursorX, cursorY = getCursorUniverseCoordinate()
		local diffX, diffY = dragStartX - cursorX, dragStartY - cursorY
		cameraX = cameraX + diffX
		cameraY = cameraY + diffY
		clampMapCamera()
		readjustCamera()
	end
end

local lastShowMinimapStyle = 0
function Cartographer.SwitchTextureStyleOnZoom(elapsed, currentTime)
	if cameraZoom > DEFAULT_ZOOM_TO_MINIMAP_TEXTURE then
		if lastShowMinimapStyle == 1 then
			return
		end
		
		addTimer(Cartographer.ShowMinimapStyleTexturesWhenInsideView)
		
		lastShowMinimapStyle = lastShowMinimapStyle + elapsed/FADE_TIME
		if lastShowMinimapStyle >= 1 then
			lastShowMinimapStyle = 1
			for zoneTexture, zoneFrame in pairs(zoneFrames) do
				if not INDOOR_CITIES[zoneTexture] then
					zoneFrame:Hide()
					if zoneFrame:GetAlpha() == 1 then
						zoneFrame:SetAlpha(0.999)
					end
				end
			end
		end
		for _, minimapStyleTextureFrame in pairs(minimapStyleTextureFrames) do
			minimapStyleTextureFrame:Show()
			minimapStyleTextureFrame:SetAlpha(lastShowMinimapStyle)
		end
	else
		if lastShowMinimapStyle == 0 then
			return
		end
		
		lastShowMinimapStyle = lastShowMinimapStyle - elapsed/FADE_TIME
		if lastShowMinimapStyle <= 0 then
			lastShowMinimapStyle = 0
			for _, minimapStyleTextureFrame in pairs(minimapStyleTextureFrames) do
				minimapStyleTextureFrame:Hide()
			end
			removeTimer(Cartographer.ShowMinimapStyleTexturesWhenInsideView)
		end
		for _, minimapStyleTextureFrame in pairs(minimapStyleTextureFrames) do
			minimapStyleTextureFrame:SetAlpha(lastShowMinimapStyle)
		end
	end
end

local lastShownTextures = {}
local tmp = {}
function Cartographer.ShowMinimapStyleTexturesWhenInsideView(elapsed, currentTime)
	lastShownTextures, tmp = tmp, lastShownTextures
	local currentContinentNum
	for continentNum, currentContinentData in ipairs(CONTINENT_DATA) do
		local continentLeft = currentContinentData[CONTINENT_DATA__SHOWN_LEFT]
		local continentRight = currentContinentData[CONTINENT_DATA__SHOWN_RIGHT]
		local continentBottom = currentContinentData[CONTINENT_DATA__SHOWN_BOTTOM]
		local continentTop = currentContinentData[CONTINENT_DATA__SHOWN_TOP]
		if cameraX >= continentLeft and cameraX <= continentRight and cameraY >= continentBottom and cameraY <= continentTop then
			currentContinentNum = continentNum
			break
		end 
	end
	if currentContinentNum then
		local currentContinentData = CONTINENT_DATA[currentContinentNum]
		local continentTexture = currentContinentData[CONTINENT_DATA__TEXTURE]
		local minimapStyleTextureFrame = minimapStyleTextureFrames[continentTexture]
		local center = cameraX - currentContinentData[CONTINENT_DATA__MAP_CENTER]
		local middle = currentContinentData[CONTINENT_DATA__MAP_MIDDLE] - cameraY
		center = center * YARDS_PER_PIXEL
		middle = middle * YARDS_PER_PIXEL
		center = math.floor(center / BLIZZARD_MINIMAP_TILE_YARD_SIZE)
		middle = math.ceil(middle / BLIZZARD_MINIMAP_TILE_YARD_SIZE)
		
		local x_len = math.ceil((scrollFrame:GetWidth()/cameraZoom * YARDS_PER_PIXEL / BLIZZARD_MINIMAP_TILE_YARD_SIZE) / 2) + 1
		local y_len = math.ceil((scrollFrame:GetHeight()/cameraZoom * YARDS_PER_PIXEL / BLIZZARD_MINIMAP_TILE_YARD_SIZE) / 2) + 1
		for i = center - x_len, center + x_len do
			for j = middle - y_len, middle + y_len do
				local tex = minimapStyleTextureFrame[i * 100 + j + 10000]
				if tex then
					tex:Show()
					lastShownTextures[tex] = true
					tmp[tex] = nil
				end
			end
		end
	end
	for tex in pairs(tmp) do
		tex:Hide()
		tmp[tex] = nil
	end
end

local lastZoneFrame
local currentZoneFrame
local recentShown = {}
function Cartographer.ShowMapTexturesWhenHovering(elapsed, currentTime)
	local fade_difference = elapsed/FADE_TIME
	local cursorZone
	if IsMouseHovering(scrollFrame) then
		cursorZone = getCursorZonePosition()
	else
		cursorZone = CURRENT_MAP_TEXTURE
	end
	local zoneFrame = zoneFrames[cursorZone]
	if zoneFrame then
		local alpha = zoneFrame:GetAlpha()
		if alpha < 1 or currentZoneFrame ~= zoneFrame then
			alpha = alpha + fade_difference
			if alpha >= 1 then
				alpha = 1
			end
			recentShown[zoneFrame] = true
			if currentZoneFrame ~= zoneFrame then
				if not currentZoneFrame or not CITIES[currentZoneFrame.texture] then
					lastZoneFrame = currentZoneFrame
				end
				currentZoneFrame = zoneFrame
			end
			currentZoneFrame:Show()
			currentZoneFrame:SetAlpha(alpha)
		end
	elseif currentZoneFrame then
		local alpha = currentZoneFrame:GetAlpha()
		if alpha < 1 then
			alpha = alpha + fade_difference
			if alpha >= 1 then
				alpha = 1
			end
			currentZoneFrame:Show()
			currentZoneFrame:SetAlpha(alpha)
		end
	end
	if lastZoneFrame then
		local alpha = lastZoneFrame:GetAlpha()
		if alpha < 1 then
			alpha = alpha + fade_difference
			if alpha >= 1 then
				alpha = 1
			end
			lastZoneFrame:Show()
			lastZoneFrame:SetAlpha(alpha)
		end
	end
	for frame in pairs(recentShown) do
		if frame ~= currentZoneFrame and frame ~= lastZoneFrame then
			local alpha = frame:GetAlpha()
			if alpha > 0 then
				alpha = alpha - fade_difference
				if alpha <= 0 then
					alpha = 0
					frame:Hide()
					recentShown[frame] = nil
				end
				frame:SetAlpha(alpha)
			end
		end
	end
end

local sliderFrame

local startGradualCameraX, startGradualCameraY, startGradualCameraZoom
local gradualCameraX, gradualCameraY, gradualCameraZoom
local gradual_isManual
local gradual_message
local function moveMap(x, y, zoom, isManual, message)
	if x or y then
		if not x or not y then
			x = x or cameraX
			y = y or cameraY
		end
	end
	if sliderFrame and zoom then
		sliderFrame.slider:set(math.log(zoom))
	end
	startGradualCameraX = cameraX
	startGradualCameraY = cameraY
	startGradualCameraZoom = cameraZoom
	gradualCameraX = x
	gradualCameraY = y
	gradualCameraZoom = zoom
	gradual_isManual = isManual
	gradual_message = message
	if not isManual then
		if zoom then
			local zoom_diff = math.abs(cameraZoom/gradualCameraZoom - 1)
			if zoom_diff < 0.01 then
				cameraZoom = zoom
				gradualCameraZoom = nil
			elseif zoom_diff < 0.2 then
				if zoom < gradualCameraZoom then
					startGradualCameraZoom = zoom * 1.2
				else
					startGradualCameraZoom = zoom / 1.2
				end
			end
		end
		if x then
			local x_diff = cameraX - x
			local y_diff = cameraY - y
			local x_y_diff = (x_diff^2 + y_diff^2)^0.5
			if x_y_diff < 1e-5 then
				cameraX = x
				gradualCameraX = nil
				cameraY = y
				gradualCameraY = nil
			elseif x_y_diff < 1 then
				local angle = math.atan2(y_diff, x_diff)
				startGradualCameraX = x + math.cos(angle)
				startGradualCameraY = y + math.sin(angle)
			end
		end
	end
end
Cartographer.moveMap = moveMap

function Cartographer.GraduallyMove(elapsed, currentTime)
	if not gradualCameraZoom and not gradualCameraX and not gradualCameraY then
		return
	end
	
	local zoomTime = gradual_isManual and MANUAL_ZOOM_TIME or ZOOM_TIME
	
	local previousCameraZoom = cameraZoom
	if gradualCameraZoom then
		cameraZoom = cameraZoom * (gradualCameraZoom / startGradualCameraZoom) ^ (elapsed / zoomTime)
		
		if (gradualCameraZoom < cameraZoom) == (gradualCameraZoom > startGradualCameraZoom) then
			cameraZoom = gradualCameraZoom
		end
	end
	
	if gradualCameraX then
		if gradualCameraZoom then
			local initialCursorX = (startGradualCameraZoom * startGradualCameraX - gradualCameraZoom * gradualCameraX) / (startGradualCameraZoom - gradualCameraZoom)
			cameraX = (cameraX - initialCursorX) * previousCameraZoom / cameraZoom + initialCursorX
		else
			cameraX = cameraX + (gradualCameraX - startGradualCameraX) * (elapsed / zoomTime)
		end
		
		if (gradualCameraX < cameraX) == (gradualCameraX > startGradualCameraX) then
			cameraX = gradualCameraX
			gradualCameraX = nil
		end
	end
	
	if gradualCameraY then
		if gradualCameraZoom then
			local initialCursorY = (startGradualCameraZoom * startGradualCameraY - gradualCameraZoom * gradualCameraY) / (startGradualCameraZoom - gradualCameraZoom)
			cameraY = (cameraY - initialCursorY) * previousCameraZoom / cameraZoom + initialCursorY
		else
			cameraY = cameraY + (gradualCameraY - startGradualCameraY) * (elapsed / zoomTime)
		end
		
		if (gradualCameraY < cameraY) == (gradualCameraY > startGradualCameraY) then
			cameraY = gradualCameraY
			gradualCameraY = nil
		end
	end
	
	if gradualCameraZoom == cameraZoom then
		gradualCameraZoom = nil
	end
	
	readjustCamera()
end

local lastToCurrentZone = true
function Cartographer.ChangeCurrentMapOnZoneHover(elapsed, currentTime)
	if IsMouseHovering(scrollFrame) then
		local cursorZone, cursorX, cursorY = getCursorZonePosition()
		if cursorZone == CURRENT_MAP_TEXTURE then
			return
		end
		local zoneData = ZONE_DATA[cursorZone]
		if zoneData then
			SetMapZoom(zoneData[ZONE_DATA__CONTINENT_NUM], zoneData[ZONE_DATA__INDEX])
			lastToCurrentZone = false
		else
			local continentData = CONTINENT_DATA[cursorZone]
			if continentData then
				SetMapZoom(continentData[CONTINENT_DATA__INDEX])
				lastToCurrentZone = false
			else
				SetMapToCurrentZone()
				lastToCurrentZone = true
			end
		end
	else
		if not lastToCurrentZone then
			SetMapToCurrentZone()
			lastToCurrentZone = true
		end
	end
end

local function AddDataToFullTooltip_unit(unit)
	local title
	if unit == "player" then
		title = L["Player:"]
	elseif unit:match("^party") then
		title = L["Party:"]
	elseif unit:match("^raid") then
		title = L["Raid:"]
	end
	local name, realm = UnitName(unit)
	if realm and realm ~= "" then
		name = name .. "-" .. realm
	end
	WorldMapTooltip:AddDoubleLine(title, name)
	
	local className, classEnglish = UnitClass(unit)
	local classColor = RAID_CLASS_COLORS[classEnglish]
	WorldMapTooltip:AddDoubleLine(L["Class:"], className, nil, nil, nil, classColor.r, classColor.g, classColor.b)
	
	local level = UnitLevel(unit)
	local levelColor = GetDifficultyColor(level)
	WorldMapTooltip:AddDoubleLine(L["Level:"], level, nil, nil, nil, levelColor.r, levelColor.g, levelColor.b)
	
	WorldMapTooltip:AddDoubleLine(L["Race:"], UnitRace(unit))
	
	local guildName, guildRank, guildRankNum = GetGuildInfo(unit)
	if guildName then
		local playerGuildName, playerGuildRank, playerGuildRankNum = GetGuildInfo('player')
		local r, g, b = 1, 1, 0
		if playerGuildName == guildName then
			r, g, b = 0, 1, 0
		end
		
		WorldMapTooltip:AddDoubleLine(L["Guild:"], ("<%s>"):format(guildName), nil, nil, nil, r, g, b)
		
		if playerGuildName == guildName then
			local rankDiff = guildRankNum - playerGuildRankNum
			if rankDiff >= 2 then
				r, g, b = 0, 1, 0
			elseif rankDiff == 1 then
				r, g, b = 0.5, 1, 0
			elseif rankDiff == 0 then
				r, g, b = 1, 1, 0
			elseif rankDiff == -1 then
				r, g, b = 1, 0.5, 0
			else
				r, g, b = 1, 0, 0
			end
		end
		WorldMapTooltip:AddDoubleLine(L["Rank:"], guildRank, nil, nil, nil, r, g, b)
		
		-- TODO: Guild note
	end
	
	-- TODO: Distance
end

local tmp = {}
local function AddDataToTooltipLine_unit(unit)
	local title
	if unit == "player" then
		title = L["Player:"]
	elseif unit:match("^party") then
		title = L["Party:"]
	elseif unit:match("^raid") then
		title = L["Raid:"]
	end
	tmp[#tmp+1] = title
	
	tmp[#tmp+1] = " "
	
	local name, realm = UnitName(unit)
	local _, class = UnitClass(unit)
	local classColor = RAID_CLASS_COLORS[class]
	tmp[#tmp+1] = "|cff"
	tmp[#tmp+1] = ("%02x"):format(classColor.r*255)
	tmp[#tmp+1] = ("%02x"):format(classColor.g*255)
	tmp[#tmp+1] = ("%02x"):format(classColor.b*255)
	tmp[#tmp+1] = name
	tmp[#tmp+1] = "|r"
	if realm and realm ~= "" then
		tmp[#tmp+1] = "-"
		tmp[#tmp+1] = realm
	end
	
	tmp[#tmp+1] = " "
	
	local level = UnitLevel(unit)
	local levelColor = GetDifficultyColor(level)
	
	tmp[#tmp+1] = "|cff"
	tmp[#tmp+1] = ("%02x"):format(levelColor.r*255)
	tmp[#tmp+1] = ("%02x"):format(levelColor.g*255)
	tmp[#tmp+1] = ("%02x"):format(levelColor.b*255)
	tmp[#tmp+1] = level
	tmp[#tmp+1] = "|r"
	
	local s = table.concat(tmp)
	for i = 1, #tmp do
		tmp[i] = nil
	end
	WorldMapTooltip:AddLine(s)
end

local function groupMemberPOI_AddDataToFullTooltip(self)
	AddDataToFullTooltip_unit(self.unit)
end
local function groupMemberPOI_AddDataToTooltipLine(self)
	AddDataToTooltipLine_unit(self.unit)
end

local nextGroupMemberColorUpdate = 0
local flashing = true
function Cartographer.ShowGroupMembers(elapsed, currentTime)
	if not CURRENT_ZONE_DATA then
		return
	end
	local prefix, amount = "raid", GetNumRaidMembers()
	local playerRaidID = 0
	if amount == 0 then
		prefix, amount = "party", GetNumPartyMembers()
	else
		playerRaidID = UnitInRaid("player") + 1
	end
	local zoneLeft = CURRENT_ZONE_DATA[ZONE_DATA__MAP_LEFT]
	local zoneWidth = CURRENT_ZONE_DATA[ZONE_DATA__MAP_WIDTH]
	local zoneBottom = CURRENT_ZONE_DATA[ZONE_DATA__MAP_BOTTOM]
	local zoneHeight = CURRENT_ZONE_DATA[ZONE_DATA__MAP_HEIGHT]
	for i = 1, amount do
		local poi = groupPOIs[i]
		if not poi then
			poi = CreateFrame("Frame", mapView:GetName() .. "_GroupMember" .. i, mapView)
			groupPOIs[i] = poi
			totalPOIs[poi] = true
			poi.AddDataToFullTooltip = groupMemberPOI_AddDataToFullTooltip
			poi.AddDataToTooltipLine = groupMemberPOI_AddDataToTooltipLine
			poi:SetWidth(1)
			poi:SetHeight(1)
			local outer = poi:CreateTexture(poi:GetName() .. "_OuterTexture", "BORDER")
			poi.outer = outer
			outer:SetAllPoints()
			outer:SetTexture([[Interface\AddOns\Cartographer3\Outer]])
			local inner = poi:CreateTexture(poi:GetName() .. "_InnerTexture", "ARTWORK")
			poi.inner = inner
			inner:SetAllPoints()
			inner:SetTexture([[Interface\AddOns\Cartographer3\Inner]])
			local text = poi:CreateTexture(poi:GetName() .. "_Text", "OVERLAY")
			poi.text = text
			text:SetAllPoints()
			text:SetVertexColor(0, 0, 0)
		end
		if i ~= playerRaidID then
			local unit = prefix .. i
			poi.unit = unit
			local x, y = GetPlayerMapPosition(unit)
			if x == 0 and y == 0 then
				poi:Hide()
			else
				poi:Show()
			
				poi:SetPoint("CENTER", x * zoneWidth + zoneLeft, (1 - y) * zoneHeight + zoneBottom)
			end
		else
			poi:Hide()
		end
	end
	for i = amount+1, #groupPOIs do
		local poi = groupPOIs[i]
		poi:Hide()
	end
	if currentTime > nextGroupMemberColorUpdate then
		nextGroupMemberColorUpdate = currentTime + 0.5
		flashing = not flashing
		for i = 1, amount do
			local poi = groupPOIs[i]
			if poi:IsShown() then
				local poi_inner = poi.inner
				local poi_outer = poi.outer
				local unit = poi.unit
				if prefix == "raid" then
					local _, _, subgroup = GetRaidRosterInfo(i)
					poi.text:SetTexture([[Interface\AddOns\Cartographer3\Group]] .. subgroup)
					poi.text:Show()
				else
					poi.text:Hide()
				end
				if flashing then
					if UnitAffectingCombat(unit) then
						poi_outer:SetVertexColor(unpack(COMBAT_COLOR))
					elseif UnitIsDeadOrGhost(unit) then
						poi_outer:SetVertexColor(unpack(DEAD_COLOR))
					elseif MapUnit_IsInactive(unit) then
						poi_outer:SetVertexColor(unpack(INACTIVE_COLOR))
					else
						poi_outer:SetVertexColor(unpack(NORMAL_STATUS_COLOR))
					end
				else
					poi_outer:SetVertexColor(unpack(NORMAL_STATUS_COLOR))
				end
				if not done then
					local _, class = UnitClass(unit)
					local classColor = RAID_CLASS_COLORS[class]
					poi_inner:SetVertexColor(classColor.r, classColor.g, classColor.b)
				end
			end
		end
	end
end

local raidMembers = {}
function Cartographer:RAID_ROSTER_UPDATE()
	local prefix, amount = "raid", GetNumRaidMembers()
	if amount == 0 then
		prefix, amount = "party", GetNumPartyMembers()
	end
	for k in pairs(raidMembers) do
		raidMembers[k] = nil
	end
	for i = 1, amount do
		local name, realm = UnitName(prefix .. i)
		if not realm then
			raidMembers[name] = true
		end
	end
end
Cartographer.PARTY_MEMBERS_CHANGED = Cartographer.RAID_ROSTER_UPDATE

local LibGuildPositions

local guildieNameToClass = setmetatable({}, {__index=function(self, name)
	for i = 1, GetNumGuildMembers() do
		local n, _, _, _, _, _, _, _, _, _, class = GetGuildRosterInfo(i)
		if n == name then
			self[name] = class
			return class
		end
	end
	self[name] = "UNKNOWN"
	return "UNKNOWN"
end})

local function guildMemberPOI_AddDataToFullTooltip(self)
	local name = self.name
	WorldMapTooltip:AddDoubleLine(L["Guild:"], name)
	
	-- TODO: add more info
end
local function guildMemberPOI_AddDataToTooltipLine(self)
	local name = self.name
	WorldMapTooltip:AddDoubleLine(L["Guild:"], name)
	
	-- TODO: add more info
end

local guildMemberNum = 0
function Cartographer:ShowGuildMember(event, name, x, y, zone)
	if not raidMembers[name] then
		local gx, gy = convertZonePositionToUniverseCoordinate(zone, x, y)
		if gx then
			local poi = guildPOIs[name]
			if not poi then
				poi = table.remove(guildPOIs)
				if not poi then
					guildMemberNum = guildMemberNum + 1
					poi = CreateFrame("Frame", mapView:GetName() .. "_GuildMember" .. guildMemberNum, mapView)
					totalPOIs[poi] = true
					poi.AddDataToFullTooltip = guildMemberPOI_AddDataToFullTooltip
					poi.AddDataToTooltipLine = guildMemberPOI_AddDataToTooltipLine
					poi:SetWidth(1)
					poi:SetHeight(1)
					local outer = poi:CreateTexture(poi:GetName() .. "_OuterTexture", "BORDER")
					poi.outer = outer
					outer:SetAllPoints()
					outer:SetTexture([[Interface\AddOns\Cartographer3\OuterGuild]])
					local inner = poi:CreateTexture(poi:GetName() .. "_InnerTexture", "ARTWORK")
					poi.inner = inner
					inner:SetAllPoints()
					inner:SetTexture([[Interface\AddOns\Cartographer3\InnerGuild]])
				end
				guildPOIs[name] = poi
				poi.name = name
				
				local class = guildieNameToClass[name]
				local classColor = RAID_CLASS_COLORS[class]
				if classColor then
					poi.inner:SetVertexColor(classColor.r, classColor.g, classColor.b)
				else
					poi.inner:SetVertexColor(0.8, 0.8, 0.8)
				end
			end
		
			poi:SetPoint("CENTER", gx, gy)
			poi:Show()
		else
			local poi = guildPOIs[name]
			if poi then
				guildPOIs[name] = nil
				guildPOIs[#guildPOIs] = poi
				poi.name = nil
				poi:Hide()
			end
		end
	end
end

local zoneSelectButton, zoomSliderButton

local function createMapHolder()
	createMapHolder = nil
	mapHolder = CreateFrame("Frame", "Cartographer_MapHolder", UIParent)
	mapHolder:SetWidth(db.profile.width)
	mapHolder:SetHeight(db.profile.height)
	mapHolder:SetPoint(db.profile.position[1], UIParent, db.profile.position[1], db.profile.position[2], db.profile.position[3])
	mapHolder:EnableMouse(true)
	mapHolder:SetMovable(true)
	mapHolder:SetResizable(true)
	mapHolder:RegisterForDrag("LeftButton")
	mapHolder:SetMinResize(MAPFRAME_MINRESIZE_WIDTH, MAPFRAME_MINRESIZE_HEIGHT)
	mapHolder:SetScript("OnShow", function(this)
		WorldMapFrame:Show()
		db.profile.shown = true
	end)
	mapHolder:SetScript("OnHide", function(this)
		WorldMapFrame:Hide()
		db.profile.shown = false
	end)
	if not db.profile.shown then
		mapHolder:Hide()
	end
	
	mapFrame = CreateFrame("Frame", "Cartographer_MapFrame", mapHolder)
	mapFrame:SetFrameLevel(mapHolder:GetFrameLevel()+2)
	mapFrame:SetAllPoints(mapHolder)
	
	mapFrame.bg = CreateFrame("Frame", mapFrame:GetName() .. "_Background", mapFrame)
	mapFrame.bg:SetAllPoints(mapFrame)
	mapFrame.bg:SetFrameLevel(mapHolder:GetFrameLevel())
	
	mapFrame.bg:SetBackdrop {
		bgFile = [[Interface\DialogFrame\UI-DialogBox-Background]],
		edgeFile = [[Interface\DialogFrame\UI-DialogBox-Border]],
		tile = true,
		tileSize = 32,
		edgeSize = 32,
		insets = {
			left = 5,
			right = 6,
			top = 5,
			bottom = 6
		}
	}
	
	positionFontString = mapHolder:CreateFontString(nil, "ARTWORK", "GameFontNormal")
	positionFontString:SetJustifyH("LEFT")
	addTimer(Cartographer.UpdatePositionFontString)
	addTimer(Cartographer.FixViewIfMoving)
	addTimer(Cartographer.ChangeCursorIfOnSide)
	mapHolder:SetScript("OnLeave", function(this)
		mouseDown = false
		SetCursor("POINT_CURSOR")
	end)
	
	mapHolder:SetScript("OnMouseDown", function(this, button)
		if button == "LeftButton" then
			mouseDown = true
		end
	end)
	mapHolder:SetScript("OnMouseUp", function(this, button)
		if button == "LeftButton" then
			mouseDown = false
		end
	end)
	mapHolder:SetScript("OnDragStart", function(this)
		moving = true
		if sizingPoint then
			SetCursor([=[Interface\Cursor\Inspect]=])
			this:StartSizing(sizingPoint)
		else
			SetCursor("POINT_CURSOR")
			this:StartMoving()
		end
	end)
	mapHolder:SetScript("OnDragStop", function(this)
		moving = false
		mouseDown = false
		SetCursor("POINT_CURSOR")
		this:StopMovingOrSizing()
		
		clampMapCamera()
		readjustCamera()
		db.profile.width = this:GetWidth()
		db.profile.height = this:GetHeight()
		
		local width, height = GetScreenWidth(), GetScreenHeight()
		local uiscale = UIParent:GetEffectiveScale()
		local scale = this:GetEffectiveScale() / uiscale
		local x, y = this:GetCenter()
		x, y = x*scale, y*scale
		local point
		if x < width/3 then
			point = "LEFT"
			x = x - this:GetWidth()/2*scale
			if x < -10*scale then
				x = -10*scale
			end
		elseif x < width*2/3 then	
			point = ""
			x = x - width/2
		else	
			point = "RIGHT"
			x = x - width + this:GetWidth()/2*scale
			if x > 10*scale then
				x = 10*scale
			end
		end
		if y < height/3 then
			point = "BOTTOM" .. point
			y = y - this:GetHeight()/2*scale
			if y < -10*scale then
				y = -10*scale
			end
		elseif y < height*2/3 then
			if point == "" then
				point = "CENTER"
			end
			y = y - height/2
		else
			point = "TOP" .. point
			y = y - height + this:GetHeight()/2*scale
			if y > 10*scale then
				y = 10*scale
			end
		end
		x, y = x/scale, y/scale
		this:ClearAllPoints()
		this:SetPoint(point, UIParent, point, x, y)
		db.profile.position[1] = point
		db.profile.position[2] = x
		db.profile.position[3] = y
	end)
	addTimer(Cartographer.ShowUIOnHover)
	
	scrollFrame = CreateFrame("ScrollFrame", mapHolder:GetName() .. "_ScrollFrame", mapHolder)
	mapView = CreateFrame("Button", mapHolder:GetName() .. "_MapView", scrollFrame)
	scrollFrame:SetScrollChild(mapView)
	scrollFrame:SetPoint("BOTTOMLEFT", mapHolder, "BOTTOMLEFT", 11, 10)
	scrollFrame:SetPoint("TOPRIGHT", mapHolder, "TOPRIGHT", -10, -30)
	mapView:SetWidth(10000)
	mapView:SetHeight(10000)
	readjustCamera()
	mapView:EnableMouse(true)
	mapView:EnableMouseWheel(true)
	mapView:RegisterForClicks("LeftButtonUp")
	mapView:SetScript("OnUpdate", _G.RequestBattlefieldPositions)
	
	positionFontString:SetPoint("BOTTOMLEFT", scrollFrame, "TOPLEFT", 0, 5)
	
	local bg = mapView:CreateTexture(mapView:GetName() .. "_Background", "BACKGROUND")
	bg:SetTexture(unpack(DEFAULT_MAPVIEW_BACKGROUND_COLOR))
	bg:SetAllPoints(mapView)
	createMapTextures()
	mapView:RegisterForDrag("LeftButton")
	addTimer(loadMapOverlayTextureSection, true)
	loadMapOverlayTextureSection()
	addTimer(loadMinimapStyleTextureSection, true)
	loadMinimapStyleTextureSection()
	addTimer(loadInstanceTextureSection, true)
	loadInstanceTextureSection()
	for zoneTexture, zoneFrame in pairs(zoneFrames) do
		if INDOOR_CITIES[zoneTexture] then
			local p = zoneFrame:GetParent()
			zoneFrame:SetParent(UIParent)
			zoneFrame:SetParent(p)
		end
	end
	playerArrow = CreateFrame("Frame", mapView:GetName() .. "_PlayerArrow", mapView)
	totalPOIs[playerArrow] = true
	playerArrow:SetWidth(20)
	playerArrow:SetHeight(20)
	playerArrow:ClearAllPoints()
	local inner = playerArrow:CreateTexture(playerArrow:GetName() .. "_InnerTexture", "ARTWORK")
	playerArrow.inner = inner
	inner:SetTexture([[Interface\AddOns\Cartographer3\InnerArrow]])
	inner:SetAllPoints(playerArrow)
	local _, playerClass = UnitClass("player")
	local classColor = RAID_CLASS_COLORS[playerClass]
	inner:SetVertexColor(classColor.r, classColor.g, classColor.b)
	local outer = playerArrow:CreateTexture(playerArrow:GetName() .. "_InnerTexture", "BORDER")
	playerArrow.outer = outer
	outer:SetTexture([[Interface\AddOns\Cartographer3\OuterArrow]])
	outer:SetAllPoints(playerArrow)
	function playerArrow:AddDataToFullTooltip()
		AddDataToFullTooltip_unit("player")
	end
	function playerArrow:AddDataToTooltipLine()
		AddDataToTooltipLine_unit("player")
	end
	addTimer(Cartographer.MovePlayerArrow)
	addTimer(Cartographer.DragMap)
	addTimer(Cartographer.SwitchTextureStyleOnZoom)
	addTimer(Cartographer.ShowMapTexturesWhenHovering)
	mapView:SetScript("OnMouseDown", function(this, button)
		if button == "LeftButton" then
			dragStartX, dragStartY = getCursorUniverseCoordinate()
		end
	end)
	mapView:SetScript("OnMouseUp", function(this, button)
		if IsMouseHovering(_G.WorldMapButton) then
			_G.WorldMapButton:GetScript("OnMouseUp")(_G.WorldMapButton, button)
		end
	end)
	mapView:SetScript("OnDragStart", function(this)
		dragging = true
		moveMap()
	end)
	mapView:SetScript("OnDragStop", function(this)
		dragging = false
	end)
	mapView:SetScript("OnDoubleClick", function(this, arg1)
		local zone = getCursorZonePosition()
		if zone then
			Cartographer.ZoomToZone(zone)
		end
	end)
	
	addTimer(Cartographer.GraduallyMove)
	mapView:SetScript("OnMouseWheel", function(this, arg1)
		local cursorX, cursorY = getCursorUniverseCoordinate()
		local zoomIn = arg1 > 0
		local zoom = gradualCameraZoom or cameraZoom
		if zoomIn then
			zoom = zoom / ZOOM_STEP
			if zoom > MAXIMUM_ZOOM then
				zoom = MAXIMUM_ZOOM
			end
		else
			zoom = zoom * ZOOM_STEP
			if zoom < MINIMUM_ZOOM then
				zoom = MINIMUM_ZOOM
			end
		end
		if zoom == cameraZoom then
			moveMap()
		else
			local cursorX, cursorY = getCursorUniverseCoordinate()
			
			moveMap(
				(cameraX - cursorX) * cameraZoom / zoom + cursorX,
				(cameraY - cursorY) * cameraZoom / zoom + cursorY,
			    zoom,
				true
			)
		end
	end)
	
	addTimer(Cartographer.ChangeCurrentMapOnZoneHover)
	
	addTimer(Cartographer.ShowGroupMembers)
	
	addTimer(Cartographer.UpdatePOITooltips)
	
	LibGuildPositions = LibStub("LibGuildPositions-1.0", true)
	if LibGuildPositions then
		LibGuildPositions.RegisterCallback(Cartographer, "Position", "ShowGuildMember")
		LibGuildPositions.RegisterCallback(Cartographer, "Clear", "ShowGuildMember")
	else
		Cartographer.ShowGuildMember = nil
	end
	
	local closeButton = CreateFrame("Button", mapFrame:GetName() .. "_CloseButton", mapFrame, "UIPanelCloseButton")
	closeButton:SetPoint("TOPRIGHT", -7, -9)
	closeButton:SetScript("OnClick", function(this)
		mapHolder:Hide()
	end)
	closeButton:GetNormalTexture():SetTexCoord(0.17, 0.8, 0.2, 0.8)
	closeButton:GetPushedTexture():SetTexCoord(0.17, 0.8, 0.2, 0.8)
	closeButton:GetHighlightTexture():SetTexCoord(0.17, 0.8, 0.2, 0.8)
	closeButton:SetWidth(22)
	closeButton:SetHeight(22)
	
	positionFontString:SetPoint("RIGHT", closeButton, "LEFT", -3, 0)
	
	local button_num = 0
	local lastLeft, lastRight
	local function makeButton(text, side, clickFunc, check, initialChecked)
		button_num = button_num + 1
		local button = CreateFrame(check and "CheckButton" or "Button", mapFrame:GetName() .. "_Button" .. button_num, mapFrame)
		if side == "RIGHT" then
			if not lastRight then
				button:SetPoint("BOTTOMRIGHT", mapFrame, "BOTTOMRIGHT", -7, 9)
			else
				button:SetPoint("RIGHT", lastRight, "LEFT", -3, 0)
			end
			lastRight = button
		else
			if not lastLeft then
				button:SetPoint("BOTTOMLEFT", mapFrame, "BOTTOMLEFT", 8, 9)
			else
				button:SetPoint("LEFT", lastLeft, "RIGHT", 3, 0)
			end
			lastLeft = button
		end
		button:SetWidth(22)
		button:SetHeight(22)
		button:EnableMouse(true)
		button:SetText(text)
		button:RegisterForClicks("LeftButtonUp")
		
		local left = button:CreateTexture(button:GetName() .. "Left", "BACKGROUND")
		left:SetTexture([[Interface\Buttons\UI-Panel-Button-Up]])
		left:SetWidth(12)
		left:SetHeight(22)
		left:SetTexCoord(0, 0.09375, 0, 0.6875)
		left:SetPoint("LEFT")
		local right = button:CreateTexture(button:GetName() .. "Right", "BACKGROUND")
		right:SetTexture([[Interface\Buttons\UI-Panel-Button-Up]])
		right:SetWidth(12)
		right:SetHeight(22)
		right:SetTexCoord(0.53125, 0.625, 0, 0.6875)
		right:SetPoint("RIGHT")
		local middle = button:CreateTexture(button:GetName() .. "Middle", "BACKGROUND")
		middle:SetTexture([[Interface\Buttons\UI-Panel-Button-Up]])
		middle:SetTexCoord(0.09375, 0.53125, 0, 0.6875)
		middle:SetPoint("LEFT", left, "RIGHT")
		middle:SetPoint("RIGHT", right, "LEFT")
		
		local highlight = button:CreateTexture(button:GetName() .. "Highlight", "HIGHLIGHT")
		highlight:SetTexture([[Interface\Buttons\UI-Panel-Button-Highlight]])
		highlight:SetTexCoord(0, 0.625, 0, 0.6875)
		highlight:SetBlendMode("ADD")
		highlight:SetAllPoints()

		button:SetHighlightTexture(highlight)
		local function setDown(this)
			button:GetFontString():SetPoint("CENTER", 1.28, -1.28)
			left:SetTexture([[Interface\Buttons\UI-Panel-Button-Down]])
			right:SetTexture([[Interface\Buttons\UI-Panel-Button-Down]])
			middle:SetTexture([[Interface\Buttons\UI-Panel-Button-Down]])
		end
		local function setUp(this)
			button:GetFontString():SetPoint("CENTER", 0, 0)
			left:SetTexture([[Interface\Buttons\UI-Panel-Button-Up]])
			right:SetTexture([[Interface\Buttons\UI-Panel-Button-Up]])
			middle:SetTexture([[Interface\Buttons\UI-Panel-Button-Up]])
		end
		button:SetScript("OnMouseDown", function(this)
			local checked = this.GetChecked and this:GetChecked()
			if checked then
				setUp(this)
			else
				setDown(this)
			end
		end)
		button:SetScript("OnMouseUp", function(this)
			local checked = this.GetChecked and this:GetChecked()
			if checked then
				setDown(this)
			else
				setUp(this)
			end
		end)
		button:SetScript("OnClick", function(this, button)
			if this.GetChecked then
				clickFunc(not not this:GetChecked())
				this:GetScript("OnMouseUp")(this)
			else
				clickFunc()
			end
		end)
		button:SetTextFontObject(GameFontNormal)
		button:SetText(text)
		button:SetPushedTextOffset(0, 0)
		if check and initialChecked then
			button:SetChecked(true)
			setDown(button)
		end
		return button
	end
	local followPlayer = db.profile.followPlayer
	local followPlayerButton = makeButton("P", "RIGHT", Cartographer.SetFollowPlayer, true, followPlayer)
	Cartographer.SetFollowPlayer(followPlayer)
	local zoomToZoneButton = makeButton("Z", "RIGHT", Cartographer.ZoomToCurrentZone, false)
	zoneSelectButton = makeButton("G", "LEFT", Cartographer.OpenGotoDropdownMenu, false)
	zoomSliderButton = makeButton("S", "RIGHT", Cartographer.OpenZoomSlider, false)
	
	WorldMapTooltip:SetParent(mapHolder)
	WorldMapTooltip:SetFrameStrata("TOOLTIP")
	WorldMapFrame:SetParent(mapView)
	WorldMapFrame:SetAlpha(0)
	WorldMapFrame:SetScale(1e-5)
	if mapHolder:IsShown() then
		WorldMapFrame:Show()
	end
	_G.WorldMapDetailFrame:SetParent(mapView)
	_G.WorldMapButton:SetParent(mapView)
	_G.WorldMapPositioningGuide:SetParent(mapView)
	_G.WorldMapDetailFrame:Show()
	_G.WorldMapButton:Show()
	_G.WorldMapPositioningGuide:Show()
	
	if UnitIsDeadOrGhost("player") then
		Cartographer:PLAYER_DEAD()
	end
end

local dropdownFrame
local gotoDropdownMenu

function Cartographer.OpenGotoDropdownMenu()
	if not dropdownFrame then
		dropdownFrame = CreateFrame("Frame", mapHolder:GetName() .. "_Dropdown", mapHolder, "UIDropDownMenuTemplate")
	end
	if not gotoDropdownMenu then
		local function menuSort(alpha, bravo)
			return alpha.text < bravo.text
		end
		gotoDropdownMenu = {}
		for i, v in ipairs(CONTINENT_DATA) do
			local t = {
				text = v[CONTINENT_DATA__LOCALIZED_NAME],
				hasArrow = true,
				func = Cartographer.ZoomToZone,
				arg1 = v[CONTINENT_DATA__TEXTURE],
				menuList = {}
			}
			local menuList = t.menuList
			local cities, zones = {}, {}
			for texture, zone in pairs(v[CONTINENT_DATA__ZONES]) do
				local x = {
					text = zone[ZONE_DATA__LOCALIZED_NAME],
					func = Cartographer.ZoomToZone,
					arg1 = texture,
				}
				if CITIES[texture] then
					cities[#cities+1] = x
				else
					zones[#zones+1] = x
				end
			end
			table.sort(cities, menuSort)
			table.sort(zones, menuSort)
			menuList[#menuList+1] = {
				text = L["Cities"],
				isTitle = true,
			}
			for i,v in ipairs(cities) do
				menuList[#menuList+1] = v
			end
			menuList[#menuList+1] = {
				text = L["Zones"],
				isTitle = true,
			}
			for i,v in ipairs(zones) do
				menuList[#menuList+1] = v
			end
			gotoDropdownMenu[i] = t
		end
		table.sort(gotoDropdownMenu, menuSort)
	end
	if DropDownList1:IsShown() and dropdownFrame.menuList == gotoDropdownMenu then
		CloseDropDownMenus()
	else
		EasyMenu(gotoDropdownMenu, dropdownFrame, zoneSelectButton, 0, 0, "MENU")
	end
end

function Cartographer.OpenZoomSlider()
	if not sliderFrame then
		sliderFrame = CreateFrame("Frame", mapFrame:GetName() .. "_SliderFrame", mapFrame)
		sliderFrame:Hide()
		
		sliderFrame:SetWidth(46)
		sliderFrame:SetHeight(170)
		sliderFrame:SetBackdrop({
			bgFile = [[Interface\Tooltips\UI-Tooltip-Background]],
			edgeFile = [[Interface\Tooltips\UI-Tooltip-Border]],
			tile = true,
			insets = {
				left = 5,
				right = 5,
				top = 5,
				bottom = 5
			},
			tileSize = 16,
			edgeSize = 16
		})
		sliderFrame:SetBackdropBorderColor(TOOLTIP_DEFAULT_COLOR.r, TOOLTIP_DEFAULT_COLOR.g, TOOLTIP_DEFAULT_COLOR.b)
		sliderFrame:SetBackdropColor(TOOLTIP_DEFAULT_BACKGROUND_COLOR.r, TOOLTIP_DEFAULT_BACKGROUND_COLOR.g, TOOLTIP_DEFAULT_BACKGROUND_COLOR.b)
		sliderFrame:EnableMouse(true)
		sliderFrame:Hide()
		sliderFrame:SetPoint("CENTER", UIParent, "CENTER")
		local slider = CreateFrame("Slider", sliderFrame:GetName() .. "_Slider", sliderFrame)
		sliderFrame.slider = slider
		slider:SetOrientation("VERTICAL")
		slider:SetMinMaxValues(0, 1)
		slider:SetValueStep(1e-10)
		slider:SetValue(0.5)
		slider:SetWidth(16)
		slider:SetHeight(128)
		slider:SetPoint("CENTER", sliderFrame, "CENTER", 0, 0)
		slider:SetBackdrop({
			bgFile = [[Interface\Buttons\UI-SliderBar-Background]],
			edgeFile = [[Interface\Buttons\UI-SliderBar-Border]],
			tile = true,
			edgeSize = 8,
			tileSize = 8,
			insets = {
				left = 3,
				right = 3,
				top = 3,
				bottom = 3
			}
		})
		slider:SetThumbTexture([[Interface\Buttons\UI-SliderBar-Button-Vertical]])
		
		local MAXIMUM_ZOOM_log = math.log(MAXIMUM_ZOOM)
		local MINIMUM_ZOOM_log = math.log(MINIMUM_ZOOM)
		local changing = false
		slider:SetScript("OnValueChanged", function(this)
			if changing then
				return
			end
			local value = (1 - this:GetValue()) * (MAXIMUM_ZOOM_log - MINIMUM_ZOOM_log) + MINIMUM_ZOOM_log
			moveMap(nil, nil, math.exp(value))
		end)
		local closeTime = 0
		sliderFrame:SetScript("OnShow", function(this)
			closeTime = GetTime() + 5
		end)
		sliderFrame:SetScript("OnUpdate", function(this)
			local currentTime = GetTime()
			if IsMouseHovering(this) then
				closeTime = currentTime + 3
			end
			if currentTime > closeTime then
				this:Hide()
			end
		end)
		
		function slider:set(value)
			changing = true
			
			self:SetValue(-(value - MINIMUM_ZOOM_log)/(MAXIMUM_ZOOM_log - MINIMUM_ZOOM_log) + 1)
			
			changing = false
		end
	end
	if sliderFrame:IsShown() then
		sliderFrame:Hide()
	else
		sliderFrame:Show()
		sliderFrame:SetPoint("BOTTOMRIGHT", zoomSliderButton, "TOPRIGHT")
	end
end

local lastPlayerX, lastPlayerY = 0, 0
function Cartographer.FollowPlayerIfMoved()
	if gradual_message ~= 'player' and IsMouseHovering(mapHolder) then
		return
	end
	local playerX, playerY = getUnitUniverseCoordinate("player")
	if not playerX or math.abs(lastPlayerX - playerX) < 1e-10 and math.abs(lastPlayerY - playerY) < 1e-10 then
		return
	end
	lastPlayerX, lastPlayerY = playerX, playerY
	Cartographer.ZoomToPlayer()
end

function Cartographer.SetFollowPlayer(follow)
	db.profile.followPlayer = follow
	if follow then
		lastPlayerX, lastPlayerY = 0, 0
		addTimer(Cartographer.FollowPlayerIfMoved)
		Cartographer.ZoomToPlayer()
	else
		if gradual_message == 'player' then
			moveMap()
		end
		removeTimer(Cartographer.FollowPlayerIfMoved)
	end
end

function Cartographer.ZoomToPlayer()
	if gradual_message == 'player' and not gradualCameraX and not gradualCameraY then
		local playerX, playerY = getUnitUniverseCoordinate("player")
		cameraX, cameraY = playerX, playerY
		readjustCamera()
		return
	end
	SetMapToCurrentZone()
	local playerX, playerY = getUnitUniverseCoordinate("player")
	if not playerX then
		playerX, playerY = 0, 0
	end
	moveMap(playerX, playerY, gradualCameraZoom, nil, 'player')
end

function Cartographer.ZoomToZone(texture)
	local zoneData = ZONE_DATA[texture]
	if not zoneData then
		local continentData = CONTINENT_DATA[texture]
		if continentData then
			local left = continentData[CONTINENT_DATA__SHOWN_LEFT]
			local right = continentData[CONTINENT_DATA__SHOWN_RIGHT]
			local bottom = continentData[CONTINENT_DATA__SHOWN_BOTTOM]
			local top = continentData[CONTINENT_DATA__SHOWN_TOP]
			moveMap((left + right) / 2, (bottom + top) / 2, math.min(scrollFrame:GetWidth() / (right - left), scrollFrame:GetHeight() / (top - bottom)))
		end
		return
	end
	local zoneLeft = zoneData[ZONE_DATA__MAP_LEFT]
	local zoneWidth = zoneData[ZONE_DATA__MAP_WIDTH]
	local zoneBottom = zoneData[ZONE_DATA__MAP_BOTTOM]
	local zoneHeight = zoneData[ZONE_DATA__MAP_HEIGHT]
	moveMap(zoneLeft + zoneWidth / 2, zoneBottom + zoneHeight / 2, math.min(scrollFrame:GetWidth() / zoneWidth, scrollFrame:GetHeight() / zoneHeight))
end

function Cartographer.ZoomToCurrentZone()
	SetMapToCurrentZone()
	Cartographer.ZoomToZone(CURRENT_MAP_TEXTURE)
end

local frame = CreateFrame("Frame", nil, UIParent)
function Cartographer:ADDON_LOADED()
	db = LibStub("AceDB-3.0"):New("Cartographer3DB", {
		profile = {
			width = DEFAULT_MAPFRAME_WIDTH,
			height = DEFAULT_MAPFRAME_HEIGHT,
			position = { "CENTER", 0, 0 },
			shown = true,
			followPlayer = true,
		}
	}, 'global')
	self.db = db
	frame:UnregisterEvent("ADDON_LOADED")
	self.ADDON_LOADED = nil
	local Cartographer3_NewOverlaysDB = _G.Cartographer3_NewOverlaysDB
	if type(Cartographer3_NewOverlaysDB) == "table" then
		for zoneTexture, overlayData in pairs(Cartographer3_NewOverlaysDB) do
			if type(overlayData) ~= "table" then
				Cartographer3_NewOverlaysDB[zoneTexture] = {}
			end
			if not OVERLAY_DATA[zoneTexture] then
				OVERLAY_DATA[zoneTexture] = {}
			end
			for tname, num in pairs(overlayData) do
				if OVERLAY_DATA[zoneTexture][tname] == num then
					overlayData[tname] = nil -- it's been upgraded in the main addon
				else
					OVERLAY_DATA[zoneTexture][tname] = num
				end
			end
			if not next(overlayData) then
				Cartographer3_NewOverlaysDB[zoneTexture] = nil
			end
		end
		if not next(Cartographer3_NewOverlaysDB) then
			_G.Cartographer3_NewOverlaysDB = nil
		end
	end
	
	initializeContinentData()
	createMapHolder()
end

function Cartographer:PLAYER_LOGIN()
	-- this is mostly so that Expo or other profiling utilities work properly
	frame:UnregisterEvent("PLAYER_LOGIN")
	self.PLAYER_LOGIN = nil
	getScaledCursorPosition = Cartographer.getScaledCursorPosition
	convertZonePositionToUniverseCoordinate = Cartographer.convertZonePositionToUniverseCoordinate
	convertUniverseCoordinateToZonePosition = Cartographer.convertUniverseCoordinateToZonePosition
	getCursorUniverseCoordinate = Cartographer.getCursorUniverseCoordinate
	getCursorZonePosition = Cartographer.getCursorZonePosition
	getUnitUniverseCoordinate = Cartographer.getUnitUniverseCoordinate
	readjustCamera = Cartographer.readjustCamera
	moveMap = Cartographer.moveMap
end

function Cartographer:UPDATE_PENDING_MAIL()
	local timeToRun = GetTime() + 1
	local function func(elapsed, currentTime)
		if currentTime < timeToRun then
			return
		end
		removeTimer(func)
		Cartographer.ZoomToCurrentZone()
	end
	addTimer(func)
	frame:UnregisterEvent("UPDATE_PENDING_MAIL")
	Cartographer.UPDATE_PENDING_MAIL = nil
end

frame:SetScript("OnEvent", function(this, event, ...)
	Cartographer[event](Cartographer, event, ...)
end)
frame:RegisterEvent("PLAYER_LOGIN")
for k, v in pairs(Cartographer) do
	if type(v) == "function" and type(k) == "string" and k:upper() == k then
		frame:RegisterEvent(k)
	end
end
local tmp = {}
frame:SetScript("OnUpdate", function(this, elapsed)
	local currentTime = GetTime()
	for timer in pairs(timers) do
		tmp[timer] = true
	end
	local mapHolder_shown = mapHolder:IsShown()
	for timer in pairs(tmp) do
		if timers[timer] and (mapHolder_shown or timers_whenNotShownAlso[timer]) then
			timer(elapsed, currentTime)
		end
		tmp[timer] = nil
	end
end)

function _G.ToggleWorldMap()
	if mapHolder:IsShown() then
		mapHolder:Hide()
	else
		mapHolder:Show()
	end
end

